#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

import copy
import logging
from dataclassy import dataclass
from enum import Enum


from zepben.evolve.model import EnergySource
from zepben.evolve.model import Breaker
from zepben.evolve.model import SinglePhaseKind
from zepben.evolve.model.phasedirection import PhaseDirection
from zepben.evolve.exceptions import PhaseException
from zepben.evolve.services.network.tracing.connectivity import get_connectivity
from zepben.evolve.exceptions import TracingException
from zepben.evolve.services.network.tracing.phase_status import normal_phases, current_phases
from zepben.evolve.services.network.tracing import queue_next_terminal
from zepben.evolve.services.network.tracing.traversals.queue import PriorityQueue
from zepben.evolve.services.network.tracing.traversals.tracing import Traversal
from zepben.evolve.services.network.tracing.phase_status import PhaseStatus
from zepben.evolve.services.network.tracing.traversals import BranchRecursiveTraversal
from zepben.evolve.services.network.tracing.util import normally_open, currently_open
from typing import Set, Callable, List, Iterable, Optional

__all__ = ["FeederProcessingStatus", "SetPhases", "FeederCbTerminalPhasesByStatus", "DelayedFeederTrace",
           "set_phases_and_queue_next", "set_current_phases_and_queue_next", "set_normal_phases_and_queue_next"]

logger = logging.getLogger("phasing.py")


class FeederProcessingStatus(Enum):
    COMPLETE = 0,
    PARTIAL = 1,
    NONE = 2


@dataclass(slots=True)
class FeederCbTerminalPhasesByStatus:
    terminal: Terminal
    in_phases: Set[SinglePhaseKind] = set()
    none_phases: Set[SinglePhaseKind] = set()
    phases_to_flow: Set[SinglePhaseKind] = set()


@dataclass(slots=True)
class DelayedFeederTrace:
    out_terminal: Terminal
    phases_to_flow: Set[SinglePhaseKind]


class SetPhases(object):
    def __init__(self):
        self.normal_traversal = BranchRecursiveTraversal(queue_next=set_normal_phases_and_queue_next,
                                                         process_queue=PriorityQueue(),
                                                         branch_queue=PriorityQueue())
        self.current_traversal = BranchRecursiveTraversal(queue_next=set_current_phases_and_queue_next,
                                                          process_queue=PriorityQueue(),
                                                          branch_queue=PriorityQueue())

    async def run(self, network: NetworkService):
        # terminals = await _apply_phases_from_feeder_cbs(network)
        await _apply_phases_from_sources(network)
        terminals = [term for es in network.objects(EnergySource) if es.num_phases() > 0 for term in es.terminals]
        if not terminals:
            raise TracingException("No feeder sources were found, tracing cannot be performed.")
        breakers = network.objects(Breaker)
        await self.run_complete(terminals, breakers)

    async def run_complete(self, terminals: Iterable[Terminal], breakers: Iterable[Breaker]):
        feeder_cbs = [br for br in breakers if br.is_substation_breaker()]
        await self._run_normal(terminals, feeder_cbs)
        await self._run_current(terminals, feeder_cbs)

    async def _run_normal(self, terminals, feeder_cbs):
        await run_set_phasing(terminals, feeder_cbs, self.normal_traversal, normally_open, normal_phases)

    async def _run_current(self, terminals, feeder_cbs):
        await run_set_phasing(terminals, feeder_cbs, self.current_traversal, currently_open, current_phases)

    async def run_ce(self, ce: ConductingEquipment, breakers: Iterable[Breaker]):
        if ce.num_terminals() == 0:
            return
        for in_term in ce.terminals:
            normal_phases_to_flow = _get_phases_to_flow(in_term, normally_open, normal_phases)
            current_phases_to_flow = _get_phases_to_flow(in_term, currently_open, current_phases)

            for out_term in ce.terminals:
                if out_term is not in_term:
                    _flow_through_equipment(self.normal_traversal, in_term, out_term, normal_phases_to_flow, normal_phases)
                    _flow_through_equipment(self.current_traversal, in_term, out_term, current_phases_to_flow, current_phases)

        self.normal_traversal.tracker.clear()
        self.current_traversal.tracker.clear()

        await self.run_complete(ce.terminals, breakers)


async def find_es_breaker_terminal(es):
    """
    From an EnergySource finds the closest connected Feeder CB (Breaker that is part of a substation).
    At the moment we assume that all EnergySource's with EnergySourcePhase's will be associated with at least a
    single feeder circuit breaker, and thus this function given an `EnergySource` will perform a trace that returns
    the first `zepben.evolve.iec61970.base.core.terminal.Terminal` encountered from that `EnergySource` that belongs to a `Breaker`. This `zepben.evolve.iec61970.base.core.terminal.Terminal` should always
    be the most downstream `zepben.evolve.iec61970.base.core.terminal.Terminal` on the `Breaker`, and thus can then be used for setting `Direction` downstream and
    away from this `Breaker`.
    TODO: check how ES are normally connected to feeder CB's.
    """
    out_terminals = set()

    async def stop_on_sub_breaker(term, exc=None):
        if out_terminals:  # stop as soon as we find a substation breaker.
            return True
        try:
            if term.conducting_equipment.is_substation_breaker():
                out_terminals.add(term)
                return True
        except AttributeError:
            return False
        return False

    t = Traversal(queue_next=queue_next_terminal, start_item=es.terminals[0], process_queue=PriorityQueue(), stop_conditions=[stop_on_sub_breaker])
    await t.trace()

    return out_terminals


async def _apply_phases_from_feeder_cbs(network):
    """
    Apply phase and direction on all Feeder Circuit Breakers. Will make all phases on the outgoing Terminal of a
    `Breaker` that is part of a substation have a `Direction` of `OUT`.
    `network` `zepben.evolve.network.Network` to apply phasing on.
    """
    start_terms = []
    # TODO: check if below assumption is correct
    # We find the substation breaker from the networks energy sources as we assume that the ES will be wired below
    # the breaker, and thus we can determine which terminal of the breaker to flow out from and apply phases.
    for es in network.energy_sources.values():
        esp = es.energy_source_phases
        if esp:
            if len(esp) != es.num_cores:
                # TODO: java network phases doesn't throw here, but would throw in the below for loop if num_phases > len(esp). why does java silently handle less cores?
                logger.error(f"Energy source {es.name} [{es.mrid}] is a source with {len(esp)} and {es.num_phases}. Number of phases should match number of cores. Phasing cannot be applied")
                raise TracingException(
                    f"Energy source {es.name} [{es.mrid}] is a source with {len(esp)} and {es.num_cores}. Number of phases should match number of cores. Phasing cannot be applied")
            breaker_terms = await find_es_breaker_terminal(es)
            for terminal in breaker_terms:
                for phase in terminal.phases.single_phases:
                    normal_phases(terminal, phase).add(esp[phase].phase, PhaseDirection.OUT)
                    current_phases(terminal, phase).add(esp[phase].phase, PhaseDirection.OUT)
                logger.debug(f"Set {terminal.conducting_equipment.mrid} as Feeder Circuit Breaker with phases {terminal.phases.phase}")
            start_terms.extend(breaker_terms)
    return start_terms


async def _apply_phases_from_sources(network: NetworkService):
    """
    Apply phase and direction on all Feeder Circuit Breakers. Will make all phases on the outgoing Terminal of a
    `Breaker` that is part of a substation have a `Direction` of `OUT`.
    `network` `zepben.evolve.network.Network` to apply phasing on.
    """
    for es in network.objects(EnergySource):
        if es.num_phases() > 0:
            await _apply_phases_from_source(es)


async def _apply_phases_from_source(energy_source: EnergySource):
    if energy_source.num_terminals() == 0:
        return
    es_phases = set()
    for phase in energy_source.phases:
        es_phases.add(phase)

    nominal_phases = set()
    for terminal in energy_source.terminals:
        nominal_phases.update(terminal.phases.single_phases)

    if len(es_phases) != len(nominal_phases):
        logger.warning((f"Energy source {str(energy_source)} is a source with {len(es_phases)} phases and {len(nominal_phases)} nominal phases. "
                        f"Number of phases should match the number of nominal phases!"))

    for term in energy_source.terminals:
        for phase in term.phases.single_phases:
            normal_phases(term, phase).add(phase, PhaseDirection.OUT)
            current_phases(term, phase).add(phase, PhaseDirection.OUT)


# TODO: pass through visited and be smart with it
def set_normal_phases_and_queue_next(terminal, traversal, visited):
    set_phases_and_queue_next(terminal, traversal, normally_open, normal_phases)


def set_current_phases_and_queue_next(terminal, traversal, visited):
    set_phases_and_queue_next(terminal, traversal, currently_open, current_phases)


def set_phases_and_queue_next(current: Terminal,
                              traversal: BranchRecursiveTraversal,
                              open_test: Callable[[ConductingEquipment, SinglePhaseKind], bool],
                              phase_selector: Callable[[Terminal, SinglePhaseKind], PhaseStatus]):
    phases_to_flow = _get_phases_to_flow(current, open_test, phase_selector)

    if current.conducting_equipment:
        for out_terminal in current.conducting_equipment.terminals:
            if out_terminal != current and _flow_through_equipment(traversal, current, out_terminal, phases_to_flow, phase_selector):
                _flow_out_to_connected_terminals_and_queue(traversal, out_terminal, phases_to_flow, phase_selector)


async def run_set_phasing(start_terminals: List[Terminal],
                          process_feeder_cbs: List[Breaker],
                          traversal: BranchRecursiveTraversal,
                          open_test: Callable[[ConductingEquipment, Optional[SinglePhaseKind]], bool],
                          phase_selector: Callable[[Terminal, Optional[SinglePhaseKind]], PhaseStatus]):
    for terminal in start_terminals:
        await _run_terminal(terminal, traversal, phase_selector)

    # We take a copy of the feeder CB's as we will modify the list while processing them.
    process_feeder_cbs = copy.copy(process_feeder_cbs)
    keep_processing = True
    while keep_processing:
        delayed_feeder_traces = []
        for feeder_cb in process_feeder_cbs:
            status = _run_feeder_breaker(feeder_cb, traversal, open_test, phase_selector, delayed_feeder_traces)
            if status == FeederProcessingStatus.COMPLETE:
                process_feeder_cbs.remove(feeder_cb)

        for trace in delayed_feeder_traces:
            await _run_from_out_terminal(traversal, trace.out_terminal, trace.phases_to_flow, phase_selector)
        keep_processing = len(delayed_feeder_traces) > 0


async def _run_terminal(start: Terminal, traversal: BranchRecursiveTraversal, phase_selector: Callable[[Terminal, SinglePhaseKind], PhaseStatus]):
    phases_to_flow = {phase for phase in start.phases.single_phases if phase_selector(start, phase).direction().has(PhaseDirection.OUT)}
    await _run_from_out_terminal(traversal, start, phases_to_flow, phase_selector)


async def _run_from_out_terminal(traversal: BranchRecursiveTraversal, out_terminal: Terminal, phases_to_flow: Set[SinglePhaseKind],
                                 phase_selector: Callable[[Terminal, SinglePhaseKind], PhaseStatus]):
    traversal.reset()
    traversal.tracker.visit(out_terminal)
    _flow_out_to_connected_terminals_and_queue(traversal, out_terminal, phases_to_flow, phase_selector)
    await traversal.trace()


def _flow_out_to_connected_terminals_and_queue(traversal: BranchRecursiveTraversal, out_terminal: Terminal, phases_to_flow: Set[SinglePhaseKind],
                                               phase_selector: Callable[[Terminal, SinglePhaseKind], PhaseStatus]):
    connectivity_results = get_connectivity(out_terminal, phases_to_flow)
    for cr in connectivity_results:
        in_term = cr.to_terminal
        has_added = False
        for oi in cr.nominal_phase_paths:
            out_core = oi.from_core
            in_core = oi.to_core
            out_phase = phase_selector(out_terminal, out_core).phase()
            in_phase = phase_selector(in_term, in_core)
            try:
                if in_phase.add(out_phase, PhaseDirection.IN):
                    has_added = True
            except PhaseException as ex:
                raise PhaseException(
                    (f"Attempted to apply more than one phase to [{in_term.conducting_equipment.mrid if in_term.conducting_equipment else in_term.mrid}"
                     f" on nominal phase {oi.to_phase}. Attempted to apply phase {out_phase} to {in_phase.phase()}."), ex)

        if has_added and not traversal.has_visited(in_term):
            if len(connectivity_results) > 1 or (out_terminal.conducting_equipment is not None and out_terminal.conducting_equipment.num_terminals() > 2):
                branch = traversal.create_branch()
                branch.start_item = in_term
                traversal.branch_queue.put(branch)
            else:
                traversal.process_queue.put(in_term)


def _get_phases_to_flow(terminal: Terminal,
                        open_test: Callable[[ConductingEquipment, Optional[SinglePhaseKind]], bool],
                        phase_selector: Callable[[Terminal, SinglePhaseKind], PhaseStatus]):
    phases_to_flow = set()
    try:
        if terminal.conducting_equipment.is_substation_breaker():
            return phases_to_flow
    except AttributeError:
        pass

    if terminal.conducting_equipment is None:
        return phases_to_flow

    equip = terminal.conducting_equipment
    for phase in terminal.phases.single_phases:
        if not open_test(equip, phase) and phase_selector(terminal, phase).direction().has(PhaseDirection.IN):
            phases_to_flow.add(phase)
    return phases_to_flow


def _flow_through_equipment(traversal: BranchRecursiveTraversal, in_terminal: Terminal, out_terminal: Terminal, phases_to_flow: Set[SinglePhaseKind],
                            phase_selector: Callable[[Terminal, SinglePhaseKind], PhaseStatus]):
    traversal.tracker.visit(out_terminal)
    has_changes = False

    for phase in phases_to_flow:
        out_phase_status = phase_selector(out_terminal, phase)
        try:
            in_phase = phase_selector(in_terminal, phase).phase()
            applied = out_phase_status.add(in_phase, PhaseDirection.OUT)
            has_changes = applied or has_changes
        except PhaseException as ex:
            raise PhaseException((
                f"Attempted to apply more than one phase to {out_terminal.conducting_equipment.mrid if out_terminal.conducting_equipment else in_terminal.mrid}"
                f" on nominal phase {phase}. Detected phases {out_phase_status.phase()} and {in_phase}. Underlying error was {str(ex)}"),
                ex)
    return has_changes


def _get_feeder_cb_terminal_cores_by_status(feeder_cb: Breaker,
                                            open_test: Callable[[ConductingEquipment, Optional[SinglePhaseKind]], bool],
                                            phase_selector: Callable[[Terminal, SinglePhaseKind], PhaseStatus]):
    res = []
    for terminal in feeder_cb.terminals:
        status = FeederCbTerminalPhasesByStatus(terminal=terminal)
        res.append(status)

        for phase in terminal.phases.single_phases:
            phase_status = phase_selector(terminal, phase)
            if phase_status.direction() == PhaseDirection.IN:
                status.in_phases.add(phase)
                if not open_test(feeder_cb, phase):
                    status.phases_to_flow.add(phase)
            elif phase_status.direction() == PhaseDirection.BOTH:
                status.in_phases.add(phase)
            elif phase_status.direction() == PhaseDirection.NONE:
                status.none_phases.add(phase)
    return res


def _flow_through_feeder_cb_and_queue(in_terminal: FeederCbTerminalPhasesByStatus,
                                      out_terminal: FeederCbTerminalPhasesByStatus,
                                      traversal: BranchRecursiveTraversal,
                                      phase_selector: Callable[[Terminal, SinglePhaseKind], PhaseStatus],
                                      delayed_traces: List,
                                      processed_phases: Set[SinglePhaseKind]):
    if not in_terminal.in_phases:
        return

    phases_to_flow = copy.copy(in_terminal.phases_to_flow)
    for phase in in_terminal.terminal.phases.single_phases:
        if phase in in_terminal.in_phases:
            processed_phases.add(phase)

            # Remove any phases that have already been processed from the other side
            if phase not in out_terminal.none_phases:
                phases_to_flow.remove(phase)

    if _flow_through_equipment(traversal, in_terminal.terminal, out_terminal.terminal, phases_to_flow, phase_selector):
        delayed_traces.append(DelayedFeederTrace(out_terminal.terminal, phases_to_flow))


def _run_feeder_breaker(feeder_cb: Breaker,
                        traversal: BranchRecursiveTraversal,
                        open_test: Callable[[ConductingEquipment, Optional[SinglePhaseKind]], bool],
                        phase_selector: Callable[[Terminal, SinglePhaseKind], PhaseStatus],
                        delayed_traces: List):
    if feeder_cb.num_terminals() not in (1, 2):
        logger.warning(f"Ignoring feeder CB {str(feeder_cb)} with {feeder_cb.num_terminals()} terminals, expected 1 or 2 terminals")
        return FeederProcessingStatus.COMPLETE

    if feeder_cb.num_terminals() == 1:
        set_phases_and_queue_next(next(feeder_cb.terminals), traversal, open_test, phase_selector)
        return FeederProcessingStatus.COMPLETE

    processed_phases = set()
    statuses = _get_feeder_cb_terminal_cores_by_status(feeder_cb, open_test, phase_selector)
    _flow_through_feeder_cb_and_queue(statuses[0], statuses[1], traversal, phase_selector, delayed_traces, processed_phases)
    _flow_through_feeder_cb_and_queue(statuses[1], statuses[0], traversal, phase_selector, delayed_traces, processed_phases)

    nominal_phases = {phase for term in feeder_cb.terminals for phase in term.phases.single_phases}
    if len(processed_phases) == len(nominal_phases):
        return FeederProcessingStatus.COMPLETE
    elif processed_phases:
        return FeederProcessingStatus.PARTIAL
    else:
        return FeederProcessingStatus.NONE



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from zepben.evolve.model.cim.iec61968.common.organisation_role import OrganisationRole

__all__ = ["AssetOrganisationRole", "AssetOwner"]


class AssetOrganisationRole(OrganisationRole):
    """ Role an organisation plays with respect to asset. """
    pass


class AssetOwner(AssetOrganisationRole):
    """ Owner of the Asset """
    pass




#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

from zepben.evolve.model.cim.iec61968.assets.asset import AssetContainer

__all__ = ["Structure"]


class Structure(AssetContainer):
    """
    Construction holding assets such as conductors, transformers, switchgear, etc. Where applicable, number of conductors
    can be derived from the number of associated wire spacing instances.
    """
    pass



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from enum import Enum


class WireMaterialKind(Enum):

    UNKNOWN = 0
    """UNKNOWN"""

    aaac = 1
    """Aluminum-alloy conductor steel reinforced."""

    acsr = 2
    """Aluminum conductor steel reinforced."""

    acsrAz = 3
    """Aluminum conductor steel reinforced, aluminumized steel core"""

    aluminum = 4
    """Aluminum wire."""

    aluminumAlloy = 5
    """Aluminum-alloy wire."""

    aluminumAlloySteel = 6
    """Aluminum-alloy-steel wire."""

    aluminumSteel = 7
    """Aluminum-steel wire."""

    copper = 8
    """Copper wire."""

    copperCadmium = 9
    """Copper cadmium wire."""

    other = 10
    """Other wire material."""

    steel = 11
    """Steel wire."""

    @property
    def short_name(self):
        return str(self)[17:]



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

from asyncio import get_event_loop
from typing import Optional, Iterable, AsyncGenerator, List, Callable

from zepben.evolve import DiagramService, IdentifiedObject
from zepben.evolve.streaming.get.consumer import CimConsumerClient, MultiObjectResult, extract_identified_object
from zepben.evolve.streaming.grpc import GrpcResult
from zepben.protobuf.dc.dc_pb2_grpc import DiagramConsumerStub
from zepben.protobuf.dc.dc_requests_pb2 import GetIdentifiedObjectsRequest

__all__ = ["DiagramConsumerClient", "SyncDiagramConsumerClient"]


class DiagramConsumerClient(CimConsumerClient):
    _stub: DiagramConsumerStub = None

    def __init__(self, channel=None, stub: DiagramConsumerStub = None, error_handlers: List[Callable[[Exception], bool]] = None):
        super().__init__(error_handlers=error_handlers)
        if channel is None and stub is None:
            raise ValueError("Must provide either a channel or a stub")
        if stub is not None:
            self._stub = stub
        else:
            self._stub = DiagramConsumerStub(channel)

    async def get_identified_object(self, service: DiagramService, mrid: str) -> GrpcResult[Optional[IdentifiedObject]]:
        """
        Retrieve the object with the given `mrid` and store the result in the `service`.

        Exceptions that occur during sending will be caught and passed to all error handlers that have been registered against this client.

        Returns a `GrpcResult` with a result of one of the following:
             - The object if found
             - None if an object could not be found or it was found but not added to `service` (see `zepben.evolve.common.base_service.BaseService.add`).
             - An `Exception` if an error occurred while retrieving or processing the object, in which case, `GrpcResult.was_successful` will return false.
        """
        async def y():
            async for io, _ in self._process_identified_objects(service, [mrid]):
                return io
            else:
                return None

        return await self.try_rpc(y)

    async def get_identified_objects(self, service: DiagramService, mrids: Iterable[str]) -> GrpcResult[MultiObjectResult]:
        """
        Retrieve the objects with the given `mrids` and store the results in the `service`.

        Exceptions that occur during sending will be caught and passed to all error handlers that have been registered against this client.

        WARNING: This operation is not atomic upon `service`, and thus if processing fails partway through `mrids`, any previously successful mRID will have been
        added to the service, and thus you may have an incomplete `BaseService`. Also note that adding to the `service` may not occur for an object if another
        object with the same mRID is already present in `service`. `MultiObjectResult.failed` can be used to check for mRIDs that were retrieved but not
        added to `service`.

        Returns a `GrpcResult` with a result of one of the following:
        - A `MultiObjectResult` containing a map of the retrieved objects keyed by mRID. If an item is not found it will be excluded from the map.
          If an item couldn't be added to `service` its mRID will be present in `MultiObjectResult.failed` (see `zepben.evolve.common.base_service.BaseService.add`).
        - An `Exception` if an error occurred while retrieving or processing the objects, in which case, `GrpcResult.was_successful` will return false.
          Note the warning above in this case.
        """
        async def y():
            results = dict()
            failed = set()
            async for io, mrid in self._process_identified_objects(service, mrids):
                if io:
                    results[io.mrid] = io
                else:
                    failed.add(mrid)
            return MultiObjectResult(results, failed)
        return await self.try_rpc(y)

    async def _process_identified_objects(self, service: DiagramService, mrids: Iterable[str]) -> AsyncGenerator[IdentifiedObject, None]:
        to_fetch = set()
        existing = set()
        for mrid in mrids:
            try:
                fetched = service.get(mrid)
                existing.add((fetched, fetched.mrid))
            except KeyError:
                to_fetch.add(mrid)

        responses = self._stub.getIdentifiedObjects(GetIdentifiedObjectsRequest(mrids=to_fetch))
        for response in responses:
            og = response.objectGroup
            io, mrid = extract_identified_object(service, og.identifiedObject)
            if io:
                yield io, mrid
            else:
                yield None, mrid
            for owned_obj in og.ownedIdentifiedObject:
                extracted, mrid = extract_identified_object(service, owned_obj)
                if extracted:
                    yield extracted, mrid
                else:
                    yield None, mrid


class SyncDiagramConsumerClient(DiagramConsumerClient):

    def get_identified_object(self, service: DiagramService, mrid: str) -> GrpcResult[Optional[IdentifiedObject]]:
        """
        Retrieve the object with the given `mrid` and store the result in the `service`.

        Exceptions that occur during sending will be caught and passed to all error handlers that have been registered against this client.

        Returns a `GrpcResult` with a result of one of the following:
             - The object if found
             - None if an object could not be found or it was found but not added to `service` (see `zepben.evolve.common.base_service.BaseService.add`).
             - An `Exception` if an error occurred while retrieving or processing the object, in which case, `GrpcResult.was_successful` will return false.
        """
        return get_event_loop().run_until_complete(super().get_identified_objects(service, mrid))

    def get_identified_objects(self, service: DiagramService, mrids: Iterable[str]) -> GrpcResult[MultiObjectResult]:
        """
        Retrieve the objects with the given `mrids` and store the results in the `service`.

        Exceptions that occur during sending will be caught and passed to all error handlers that have been registered against this client.

        WARNING: This operation is not atomic upon `service`, and thus if processing fails partway through `mrids`, any previously successful mRID will have been
        added to the service, and thus you may have an incomplete `BaseService`. Also note that adding to the `service` may not occur for an object if another
        object with the same mRID is already present in `service`. `MultiObjectResult.failed` can be used to check for mRIDs that were retrieved but not
        added to `service`.

        Returns a `GrpcResult` with a result of one of the following:
        - A `MultiObjectResult` containing a map of the retrieved objects keyed by mRID. If an item is not found it will be excluded from the map.
          If an item couldn't be added to `service` its mRID will be present in `MultiObjectResult.failed` (see `zepben.evolve.common.base_service.BaseService.add`).
        - An `Exception` if an error occurred while retrieving or processing the objects, in which case, `GrpcResult.was_successful` will return false.
          Note the warning above in this case.
        """
        return get_event_loop().run_until_complete(super().get_identified_objects(service, mrids))



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from zepben.protobuf.cim.iec61970.base.core.IdentifiedObject_pb2 import IdentifiedObject as PBIdentifiedObject
from zepben.protobuf.cim.iec61968.assetinfo.CableInfo_pb2 import CableInfo as PBCableInfo
from zepben.protobuf.cim.iec61968.assetinfo.OverheadWireInfo_pb2 import OverheadWireInfo as PBOverheadWireInfo
from zepben.protobuf.cim.iec61968.assetinfo.WireInfo_pb2 import WireInfo as PBWireInfo
from zepben.protobuf.cim.iec61968.assetinfo.WireMaterialKind_pb2 import WireMaterialKind as PBWireMaterialKind
from zepben.protobuf.cim.iec61968.assets.AssetInfo_pb2 import AssetInfo as PBAssetInfo
from zepben.protobuf.cim.iec61968.metering.EndDevice_pb2 import EndDevice as PBEndDevice
from zepben.protobuf.cim.iec61968.metering.Meter_pb2 import Meter as PBMeter
from zepben.protobuf.cim.iec61968.metering.UsagePoint_pb2 import UsagePoint as PBUsagePoint
from zepben.protobuf.cim.iec61970.base.core.AcDcTerminal_pb2 import AcDcTerminal as PBAcDcTerminal
from zepben.protobuf.cim.iec61970.base.core.BaseVoltage_pb2 import BaseVoltage as PBBaseVoltage
from zepben.protobuf.cim.iec61970.base.core.ConductingEquipment_pb2 import ConductingEquipment as PBConductingEquipment
from zepben.protobuf.cim.iec61970.base.core.Equipment_pb2 import Equipment as PBEquipment
from zepben.protobuf.cim.iec61970.base.core.Terminal_pb2 import Terminal as PBTerminal
from zepben.protobuf.cim.iec61970.base.wires.AcLineSegment_pb2 import AcLineSegment as PBAcLineSegment
from zepben.protobuf.cim.iec61970.base.wires.Breaker_pb2 import Breaker as PBBreaker
from zepben.protobuf.cim.iec61970.base.wires.Conductor_pb2 import Conductor as PBConductor
from zepben.protobuf.cim.iec61970.base.wires.EnergyConnection_pb2 import EnergyConnection as PBEnergyConnection
from zepben.protobuf.cim.iec61970.base.wires.EnergySource_pb2 import EnergySource as PBEnergySource
from zepben.protobuf.cim.iec61970.base.wires.PerLengthImpedance_pb2 import PerLengthImpedance as PBPerLengthImpedance
from zepben.protobuf.cim.iec61970.base.wires.PerLengthLineParameter_pb2 import \
    PerLengthLineParameter as PBPerLengthLineParameter
from zepben.protobuf.cim.iec61970.base.wires.PerLengthSequenceImpedance_pb2 import \
    PerLengthSequenceImpedance as PBPerLengthSequenceImpedance
from zepben.protobuf.cim.iec61970.base.wires.PowerTransformer_pb2 import PowerTransformer as PBPowerTransformer
from zepben.protobuf.cim.iec61970.base.wires.ProtectedSwitch_pb2 import ProtectedSwitch as PBProtectedSwitch
from zepben.protobuf.cim.iec61970.base.wires.Switch_pb2 import Switch as PBSwitch

from zepben.evolve.services.network.network import NetworkService

from zepben.evolve.services.network.translator.network_proto2cim import NetworkProtoToCim


class TestNetworkToCim(object):
    def test_add_pb(self):
        """Test addition to the network works for CableInfo PB type."""
       
        # Create network
        network = NetworkProtoToCim(NetworkService())

        # BaseVoltage1
        io_bv1 = PBIdentifiedObject(mRID="bv1", name="bv1")
        bv1 = PBBaseVoltage(io=io_bv1, nominalVoltage=22000)
        network.add_from_pb(bv1)

        # BaseVoltage2
        io_bv2 = PBIdentifiedObject(mRID="bv2", name="bv2")
        bv2 = PBBaseVoltage(io=io_bv2, nominalVoltage=415)
        network.add_from_pb(bv2)

        # Terminal
        io_ad = PBIdentifiedObject(mRID="t", name="t")
        ad = PBAcDcTerminal(io=io_ad, connected=True)
        t = PBTerminal(ad=ad, connectivityNodeMRID="c1")
        eq = PBEquipment(inService=True)
        ce1 = PBConductingEquipment(eq=eq, baseVoltageMRID="bv1", terminalMRIDs=["t"])
        ec = PBEnergyConnection(ce=ce1)
        es = PBEnergySource(ec=ec)
        # AttributeError: 'EnergySource' object has no attribute 'conducting_equipment'
        network.add_from_pb(es)

        # PerLengthSequenceImpedance
        io_lp = PBIdentifiedObject(mRID="plsi1", name="plsi1")
        lp = PBPerLengthLineParameter(io=io_lp)
        pli = PBPerLengthImpedance(lp=lp)
        plsi = PBPerLengthSequenceImpedance(pli=pli)
        # PerLengthLineParameter.mrid = lambda self: self.lo.mRID
        # E   AttributeError: lo
        #network.add_from_pb(plsi)

        # CableInfo
        io_ci = PBIdentifiedObject(mRID="7", name="ci")
        ai = PBAssetInfo(io=io_ci)
        wi = PBWireInfo(ai=ai, ratedCurrent=12, material=PBWireMaterialKind.aaac)
        ci = PBCableInfo(wi=wi)
        network.add_from_pb(ci)

        # AcLineSegment
        ce2 = PBConductingEquipment(baseVoltageMRID="bv1")
        cd = PBConductor(ce=ce2)
        acls = PBAcLineSegment(cd=cd, perLengthSequenceImpedanceMRID="plsi1")
        # PerLengthLineParameter.mrid = lambda self: self.lo.mRID
        # E   AttributeError: lo
        # network.add_from_pb(acls)

        # PowerTransformer
        ce3 = PBConductingEquipment(baseVoltageMRID="bv2")
        pt = PBPowerTransformer(ce=ce3)
        network.add_from_pb(pt)

        # Breaker
        ce4 = PBConductingEquipment(baseVoltageMRID="bv2")
        sw = PBSwitch(ce=ce4)
        psw = PBProtectedSwitch(sw=sw)
        br = PBBreaker(sw=psw)
        network.add_from_pb(br)

        # DOES THIS EXIST?
        # cust = PBCustomer()

        # OverheadWireInfo
        io_owi = PBIdentifiedObject(mRID="8", name="owi")
        ai2 = PBAssetInfo(io=io_owi)
        wi2 = PBWireInfo(ai=ai2)
        owi = PBOverheadWireInfo(wi=wi2)
        network.add_from_pb(owi)

        # UsagePoint
        io_up = PBIdentifiedObject(mRID="up1", name="up")
        up = PBUsagePoint(io=io_up)
        network.add_from_pb(up)

        # Meter
        ed = PBEndDevice(usagePointMRIDs=["up1"])
        me = PBMeter(ed=ed)
        network.add_from_pb(me)

        print(network.service)


#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from zepben.protobuf.cim.iec61970.base.core.IdentifiedObject_pb2 import IdentifiedObject as PBIdentifiedObject

__all__ = ["mrid_or_empty"]


def mrid_or_empty(io: PBIdentifiedObject) -> str:
    return str(io.mrid) if io else ""



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from dataclassy import dataclass

from zepben.evolve.services.common import BaseService

__all__ = ["CustomerService"]


class CustomerService(BaseService):
    """
    Used to store Customer related types.
    """
    name: str = "customer"



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

import zepben.evolve.streaming as streaming
import asyncio
import grpc 


class TestStreaming(object):
    pass
    # def test_retrieve_network(self):
    #     """Test retrieve_network"""
    #     channel = grpc.insecure_channel('localhost:50051')
    #     network = asyncio.run(streaming.retrieve_network(channel))
    #     print(len(network._unresolved_references))


#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from abc import abstractmethod

__all__ = ["BaseTracker", "Tracker"]

from typing import Set

from dataclassy import dataclass




@dataclass(slots=True)
class BaseTracker(object):
    """
    An interface used by `zepben.evolve.tracing.Traversal`'s to 'track' items that have been visited.

    A `Traversal` will utilise `has_visited`, `visit`, and `clear`.
    """

    @abstractmethod
    def has_visited(self, item):
        """
        Check if the tracker has already seen an item.
        `item` The item to check if it has been visited.
        Returns true if the item has been visited, otherwise false.
        """
        raise NotImplementedError()

    @abstractmethod
    def visit(self, item):
        """
        Visit an item. Item will not be visited if it has previously been visited.
        `item` The item to visit.
        Returns True if visit succeeds. False otherwise.
        """
        raise NotImplementedError()

    @abstractmethod
    def clear(self):
        """
        Clear the tracker, removing all visited items.
        """
        raise NotImplementedError()


class Tracker(BaseTracker):
    """
    An interface used by `zepben.evolve.traversals.tracing.Traversal`'s to 'track' items that have been visited.
    """
    visited: Set = set()

    def has_visited(self, item):
        """
        Check if the tracker has already seen an item.
        `item` The item to check if it has been visited.
        Returns true if the item has been visited, otherwise false.
        """
        return item in self.visited

    def visit(self, item):
        """
        Visit an item. Item will not be visited if it has previously been visited.
        `item` The item to visit.
        Returns True if visit succeeds. False otherwise.
        """
        if item in self.visited:
            return False
        else:
            self.visited.add(item)
            return True

    def clear(self):
        """
        Clear the tracker, removing all visited items.
        """
        self.visited.clear()

    def copy(self):
        return Tracker(visited=self.visited.copy())



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

from zepben.evolve.model import SinglePhaseKind
from zepben.evolve.model.phasedirection import PhaseDirection
from abc import ABC, abstractmethod

__all__ = ["normal_phases", "current_phases", "PhaseStatus", "NormalPhases", "CurrentPhases"]


def normal_phases(terminal: Terminal, phase: SinglePhaseKind):
    return NormalPhases(terminal, phase)


def current_phases(terminal: Terminal, phase: SinglePhaseKind):
    return CurrentPhases(terminal, phase)


class PhaseStatus(ABC):

    @abstractmethod
    def phase(self):
        """
        Returns The phase added to this status
        """
        raise NotImplementedError()

    @abstractmethod
    def direction(self):
        """
        Returns The direction added to this status.
        """
        raise NotImplementedError()

    @abstractmethod
    def set(self, phase: SinglePhaseKind, direction: PhaseDirection):
        """
        Clears the phase and sets it to the specified phase and direction.
        If the passed in phase is NONE or the passed in direction is NONE, this should clear the phase status.
        `phase` The new phase to be set.
        `direction` The direction of the phase.
        """
        raise NotImplementedError()

    @abstractmethod
    def add(self, phase: SinglePhaseKind, direction: PhaseDirection):
        """
        Adds a phase to the status with the given direction.
        `phase` The phase to be added.
        `direction` The direction of the phase.
        Returns True if the phase or direction has been updated
        """
        raise NotImplementedError()

    @abstractmethod
    def remove(self, phase: SinglePhaseKind, direction: PhaseDirection = None):
        """
        Removes a phase from the status. If direction is supplied will remove phase matching the direction.
        `phase` The phase to be removed.
        `direction` The direction to match with the phase being removed.
        Returns True if the phase or direction has been removed.
        """
        raise NotImplementedError()


class NormalPhases(PhaseStatus):

    def __init__(self, terminal: Terminal, nominal_phase: SinglePhaseKind):
        self.terminal = terminal
        self.nominal_phase = nominal_phase

    def phase(self):
        return self.terminal.traced_phases.phase_normal(self.nominal_phase)

    def direction(self):
        return self.terminal.traced_phases.direction_normal(self.nominal_phase)

    def set(self, phase_kind: SinglePhaseKind, dir: PhaseDirection):
        return self.terminal.traced_phases.set_normal(phase_kind, self.nominal_phase, dir)

    def add(self, phase_kind: SinglePhaseKind, dir: PhaseDirection):
        return self.terminal.traced_phases.add_normal(phase_kind, self.nominal_phase, dir)

    def remove(self, phase_kind: SinglePhaseKind, dir: PhaseDirection = None):
        if dir is not None:
            return self.terminal.traced_phases.remove_normal(phase_kind, self.nominal_phase, dir)
        else:
            return self.terminal.traced_phases.remove_normal(phase_kind, self.nominal_phase)


class CurrentPhases(PhaseStatus):

    def __init__(self, terminal: Terminal, nominal_phase: SinglePhaseKind):
        self.terminal = terminal
        self.nominal_phase = nominal_phase

    def phase(self):
        return self.terminal.traced_phases.phase_current(self.nominal_phase)

    def direction(self):
        return self.terminal.traced_phases.direction_current(self.nominal_phase)

    def set(self, phase_kind: SinglePhaseKind, dir: PhaseDirection):
        return self.terminal.traced_phases.set_current(phase_kind, self.nominal_phase, dir)

    def add(self, phase_kind: SinglePhaseKind, dir: PhaseDirection):
        return self.terminal.traced_phases.add_current(phase_kind, self.nominal_phase, dir)

    def remove(self, phase_kind: SinglePhaseKind, dir: PhaseDirection = None):
        if dir is not None:
            return self.terminal.traced_phases.remove_current(phase_kind, self.nominal_phase, dir)
        else:
            return self.terminal.traced_phases.remove_current(phase_kind, self.nominal_phase)




#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations
import logging
from zepben.evolve.services.network.tracing.traversals.tracing import Traversal
from zepben.evolve.services.network.tracing.traversals.queue import LifoQueue
from zepben.evolve.services.network.tracing.phase_status import normal_phases, current_phases

__all__ = ["normally_open", "currently_open", "ignore_open", "phase_log"]
phase_logger = logging.getLogger("phase_logger")


def normally_open(equip: ConductingEquipment, phase: Optional[SinglePhaseKind] = None):
    """
    Test if a given phase on an equipment is normally open.
    `equip` The equipment to test
    `phase` The Phase to test. If None tests all phases.
    Returns True if the equipment is open (de-energised), False otherwise
    """
    try:
        return not equip.normally_in_service or equip.is_normally_open(phase)
    except AttributeError:
        return not equip.normally_in_service


def currently_open(equip: ConductingEquipment, phase: Optional[SinglePhaseKind] = None):
    """
    Test if a given phase on an equipment is open.
    `equip` The equipment to test
    `phase` The phase to test. If None tests all phases.
    Returns True if the equipment is open (de-energised), False otherwise
    """
    try:
        return not equip.in_service or equip.is_open(phase)
    except AttributeError:
        return not equip.in_service


def ignore_open(ce: ConductingEquipment, phase: Optional[SinglePhaseKind] = None):
    """
    Test that always returns that the phase is closed.
    `equip` The equipment to test
    `phase` The phase to test. If None tests all cores.
    Returns False 
    """
    return False


async def phase_log(cond_equip):
    msg = ""
    try:
        for e in cond_equip:
            msg = await _phase_log_trace(e)
    except:
        msg = await _phase_log_trace(cond_equip)
    phase_logger.debug(msg)


async def _phase_log_trace(cond_equip):
    log_msg = []

    async def log(e, exc):
        equip_msgs = []
        for term in e.terminals:
            e_msg = f"{e.mrid}-T{term.sequence_number}:"
            for n in term.phases.single_phases:
                ps_n = normal_phases(term, n)
                phase_n_msg = f"n: {ps_n.phase().short_name}:{ps_n.direction().short_name}"
                ps_c = current_phases(term, n)
                phase_c_msg = f"c: {ps_c.phase().short_name}:{ps_c.direction().short_name}"
                e_msg = f"{e_msg} {{core {n}: {phase_n_msg} {phase_c_msg}}}"
            equip_msgs.append(e_msg)
        log_msg.append(equip_msgs)

    trace = Traversal(queue_next=queue_next_equipment, start_item=cond_equip, process_queue=LifoQueue(), step_actions=[log])
    await trace.trace()
    return "\n".join([", ".join(x) for x in log_msg])


def queue_next_equipment(item, exclude=None):
    connected_equips = item.get_connected_equipment(exclude=exclude)
    tracing_logger.debug(f"Queuing connections [{', '.join(e.mrid for e in connected_equips)}] from {item.mrid}")
    return connected_equips



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from zepben.evolve.services.network.tracing.traversals.queue import Queue
from zepben.evolve.services.network.tracing.traversals.tracing import BaseTraversal
from zepben.evolve.services.network.tracing.traversals import Tracker
from typing import Callable, Set, TypeVar, Optional

__all__ = ["BranchRecursiveTraversal"]
T = TypeVar('T')


class BranchRecursiveTraversal(BaseTraversal[T]):

    queue_next: Callable[[T, BaseTraversal[T], Set[T]], None]
    """A callable for each item encountered during the trace, that should queue the next items found on the given traversal's `process_queue`. 
    The first argument will be the current item, the second this traversal, and the third a set of already visited items that can be used as an optional 
    optimisation to skip queuing."""

    branch_queue: Queue
    """Queue containing branches to be processed"""

    process_queue: Queue
    """Queue containing the items to process for this branch"""

    tracker: Tracker = Tracker()
    """Tracker for the items in this branch"""

    parent: Optional[BaseTraversal] = None
    """The parent branch for this branch, None implies this branch has no parent"""

    on_branch_start: Optional[Callable[[T], None]] = None
    """A function to call at the start of each branches processing"""

    def __lt__(self, other):
        """
        This Traversal is Less than `other` if the starting item is less than other's starting item.
        This is used to dictate which branch is next to traverse in the branch_queue.
        """
        if self.start_item is not None and other.start_item is not None:
            return self.start_item < other.start_item
        elif self.start_item is None and other.start_item is None:
            return False
        elif other.start_item is None:
            return True
        else:
            return False

    def has_visited(self, item: T):
        """
        Check whether item has been visited before. An item is visited if this traversal or any parent has visited it.

        `item` The item to check
        Returns True if the item has been visited once.
        """
        parent = self.parent
        while parent is not None:
            if parent.tracker.has_visited(item):
                return True
            parent = parent.parent

        return self.tracker.has_visited(item)

    def visit(self, item: T):
        """
        Visit an item.
        `item` Item to visit
        Returns True if we visit the item. False if this traversal or any parent has previously visited this item.
        """
        parent = self.parent
        while parent is not None:
            if parent.tracker.has_visited(item):
                return False
            parent = parent.parent
        return self.tracker.visit(item)

    async def traverse_branches(self):
        """
        Start a new traversal for the next branch in the queue.
        on_branch_start will be called on the start_item for the branch.
        """
        while not self.branch_queue.empty():
            t = self.branch_queue.get()
            if t is not None:
                if self.on_branch_start is not None:
                    self.on_branch_start(t.start_item)
                await t.trace()

    def reset(self):
        """Reset the run state, queues and tracker for this this traversal"""
        self._reset_run_flag()
        self.process_queue.queue.clear()
        self.branch_queue.queue.clear()
        self.tracker.clear()

    def create_branch(self):
        """
        Create a branch for this `Traversal`. Will take copies of queues, actions, conditions, and tracker, and
        pass this `Traversal` as the parent. The new Traversal will be :meth:`reset` prior to being returned.
        Returns A new `BranchRecursiveTraversal` the same as this, but with this Traversal as its parent
        """
        branch = BranchRecursiveTraversal(queue_next=self.queue_next,
                                          branch_queue=self.branch_queue.copy(),
                                          tracker=self.tracker.copy(),
                                          parent=self,
                                          on_branch_start=self.on_branch_start,
                                          process_queue=self.process_queue.copy(),
                                          step_actions=list(self.step_actions),
                                          stop_conditions=list(self.stop_conditions))
        branch.reset()
        return branch

    async def _run_trace(self, can_stop_on_start_item: bool = True):
        """
        Run's the trace. Stop conditions and step_actions are called with await, so you can utilise asyncio when performing a trace if your step actions or
        conditions are IO intensive. Stop conditions and step actions will always be called for each item in the order provided.
        `can_stop_on_start_item` Whether the trace can stop on the start_item. Actions will still be applied to the start_item.
        """
        # Unroll first iteration of loop to handle can_stop_on_start_item = True
        if self.start_item is None:
            try:
                self.start_item = self.process_queue.get()
            except IndexError:
                # Our start point may very well be a branch - if so we don't need to process this branch.
                await self.traverse_branches()
                return

        self.tracker.visit(self.start_item)
        # If we can't stop on the start item we don't run any stop conditions. if this causes a problem for you,
        # work around it by running the stop conditions for the start item prior to running the trace.
        stopping = can_stop_on_start_item and await self.matches_stop_condition(self.start_item)
        await self.apply_step_actions(self.start_item, stopping)
        if not stopping:
            self.queue_next(self.start_item, self, self.tracker.visited)

        while not self.process_queue.empty():
            current = self.process_queue.get()
            if self.visit(current):
                stopping = await self.matches_stop_condition(current)
                await self.apply_step_actions(current, stopping)
                if not stopping:
                    self.queue_next(current, self, self.tracker.visited)

        await self.traverse_branches()



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from zepben.evolve.model import OrganisationRole, Asset, Pole, Streetlight, Customer, CustomerAgreement, PricingStructure, EndDevice, UsagePoint, \
    OperationalRestriction, AuxiliaryEquipment, ConductingEquipment, ConnectivityNode, Equipment, Terminal, Diagram, DiagramObject, Measurement, Control, \
    RemoteSource, RemoteControl, Conductor, EnergyConsumer, EnergyConsumerPhase, EnergySource, EnergySourcePhase, Circuit, Loop, Feeder, EquipmentContainer, \
    Substation, GeographicalRegion, SubGeographicalRegion, PowerSystemResource, PowerTransformer, PowerTransformerEnd, RatioTapChanger, TransformerEnd
from zepben.evolve.services.common.reference_resolvers import *

__all__ = ["per_length_sequence_impedance", "organisation_roles", "at_location", "ae_terminal", "ce_base_voltage", "ce_terminals",
           "asset_info", "streetlights", "pole", "cn_terminals", "remote_control", "agreements", "customer",
           "pricing_structures",
           "diagram_objects", "diagram", "service_location", "ed_usage_points", "containers", "current_feeders",
           "operational_restrictions",
           "eq_usage_points", "ec_equipment", "ec_phases", "energy_consumer", "es_phases", "energy_source", "current_equipment",
           "normal_energizing_substation", "normal_head_terminal", "sub_geographical_regions", "remote_source",
           "or_equipment", "organisation",
           "psr_location", "ends", "power_transformer", "tariffs", "transformer_end", "control", "measurement",
           "geographical_region", "substations",
           "normal_energizing_feeders", "sub_geographical_region", "conducting_equipment", "connectivity_node",
           "te_base_voltage", "ratio_tap_changer",
           "te_terminal", "end_devices", "up_equipment", "usage_point_location"]


def per_length_sequence_impedance(aclinesegment):
    return BoundReferenceResolver(aclinesegment, acls_to_plsi_resolver, None)


def organisation_roles(asset: Asset) -> BoundReferenceResolver:
    return BoundReferenceResolver(asset, asset_to_asset_org_role_resolver, None)


def at_location(asset: Asset) -> BoundReferenceResolver:
    return BoundReferenceResolver(asset, asset_to_location_resolver, None)


def ae_terminal(auxiliaryEquipment: AuxiliaryEquipment) -> BoundReferenceResolver:
    return BoundReferenceResolver(auxiliaryEquipment, aux_equip_to_term_resolver, None)


def ce_base_voltage(conducting_equipment: ConductingEquipment) -> BoundReferenceResolver:
    return BoundReferenceResolver(conducting_equipment, cond_equip_to_bv_resolver, None)


def ce_terminals(conducting_equipment: ConductingEquipment) -> BoundReferenceResolver:
    return BoundReferenceResolver(conducting_equipment, cond_equip_to_terminal_resolver, term_to_ce_resolver)


def asset_info(conductor: Conductor) -> BoundReferenceResolver:
    return BoundReferenceResolver(conductor, conductor_to_wire_info_resolver, None)


def streetlights(pole: Pole) -> BoundReferenceResolver:
    return BoundReferenceResolver(pole, pole_to_streetlight_resolver, streetlight_to_pole_resolver)


def pole(streetlight: Streetlight) -> BoundReferenceResolver:
    return BoundReferenceResolver(streetlight, streetlight_to_pole_resolver, pole_to_streetlight_resolver)


def cn_terminals(connectivity_node: ConnectivityNode) -> BoundReferenceResolver:
    return BoundReferenceResolver(connectivity_node, conn_node_to_term_resolver, term_to_cn_resolver)


def remote_control(control: Control) -> BoundReferenceResolver:
    return BoundReferenceResolver(control, control_to_remote_control_resolver, rc_to_cont_resolver)


def agreements(customer: Customer) -> BoundReferenceResolver:
    return BoundReferenceResolver(customer, cust_to_custagr_resolver, custagr_to_cust_resolver)


def customer(customer_agreement: CustomerAgreement) -> BoundReferenceResolver:
    return BoundReferenceResolver(customer_agreement, custagr_to_cust_resolver, cust_to_custagr_resolver)


def pricing_structures(customer_agreement: CustomerAgreement) -> BoundReferenceResolver:
    return BoundReferenceResolver(customer_agreement, custagr_to_ps_resolver, None)


def diagram_objects(diagram: Diagram) -> BoundReferenceResolver:
    return BoundReferenceResolver(diagram, diag_to_diagobj_resolver, diagobj_to_diag_resolver)


def diagram(diagram_object: DiagramObject) -> BoundReferenceResolver:
    return BoundReferenceResolver(diagram_object, diagobj_to_diag_resolver, diag_to_diagobj_resolver)


def service_location(end_device: EndDevice) -> BoundReferenceResolver:
    return BoundReferenceResolver(end_device, ed_to_loc_resolver, None)


def ed_usage_points(end_device: EndDevice) -> BoundReferenceResolver:
    return BoundReferenceResolver(end_device, ed_to_up_resolver, up_to_ed_resolver)


def containers(equipment: Equipment) -> BoundReferenceResolver:
    return BoundReferenceResolver(equipment, eq_to_ec_resolver, ec_to_eq_resolver)


def current_feeders(equipment: Equipment) -> BoundReferenceResolver:
    return BoundReferenceResolver(equipment, eq_to_curfeeder_resolver, curfeeder_to_eq_resolver)


def operational_restrictions(equipment: Equipment) -> BoundReferenceResolver:
    return BoundReferenceResolver(equipment, eq_to_or_resolver, or_to_eq_resolver)


def eq_usage_points(equipment: Equipment) -> BoundReferenceResolver:
    return BoundReferenceResolver(equipment, eq_to_up_resolver, up_to_eq_resolver)


def ec_equipment(equipment_container: EquipmentContainer) -> BoundReferenceResolver:
    return BoundReferenceResolver(equipment_container, ec_to_eq_resolver, eq_to_ec_resolver)


def ec_phases(energy_consumer: EnergyConsumer) -> BoundReferenceResolver:
    return BoundReferenceResolver(energy_consumer, ec_to_ecp_resolver, ecp_to_ec_resolver)


def energy_consumer(energy_consumer_phase: EnergyConsumerPhase) -> BoundReferenceResolver:
    return BoundReferenceResolver(energy_consumer_phase, ecp_to_ec_resolver, ec_to_ecp_resolver)


def es_phases(energy_source: EnergySource) -> BoundReferenceResolver:
    return BoundReferenceResolver(energy_source, es_to_esp_resolver, esp_to_es_resolver)


def energy_source(energy_source_phase: EnergySourcePhase) -> BoundReferenceResolver:
    return BoundReferenceResolver(energy_source_phase, esp_to_es_resolver, es_to_esp_resolver)


def current_equipment(feeder: Feeder) -> BoundReferenceResolver:
    return BoundReferenceResolver(feeder, curfeeder_to_eq_resolver, eq_to_curfeeder_resolver)


def normal_energizing_substation(feeder: Feeder) -> BoundReferenceResolver:
    return BoundReferenceResolver(feeder, feeder_to_nes_resolver, sub_to_feeder_resolver)


def normal_head_terminal(feeder: Feeder) -> BoundReferenceResolver:
    return BoundReferenceResolver(feeder, feeder_to_nht_resolver, None)


def sub_geographical_regions(geographical_region: GeographicalRegion) -> BoundReferenceResolver:
    return BoundReferenceResolver(geographical_region, gr_to_sgr_resolver, sgr_to_gr_resolver)


def remote_source(measurement: Measurement) -> BoundReferenceResolver:
    return BoundReferenceResolver(measurement, meas_to_rs_resolver, rs_to_meas_resolver)


def or_equipment(operational_restriction: OperationalRestriction) -> BoundReferenceResolver:
    return BoundReferenceResolver(operational_restriction, or_to_eq_resolver, eq_to_or_resolver)


def organisation(organisation_role: OrganisationRole) -> BoundReferenceResolver:
    return BoundReferenceResolver(organisation_role, orgr_to_org_resolver, None)


def psr_location(power_system_resource: PowerSystemResource) -> BoundReferenceResolver:
    return BoundReferenceResolver(power_system_resource, psr_to_loc_resolver, None)


def ends(power_transformer: PowerTransformer) -> BoundReferenceResolver:
    return BoundReferenceResolver(power_transformer, pt_to_pte_resolver, pte_to_pt_resolver)


def power_transformer(power_transformerEnd: PowerTransformerEnd) -> BoundReferenceResolver:
    return BoundReferenceResolver(power_transformerEnd, pte_to_pt_resolver, pt_to_pte_resolver)


def tariffs(pricing_structure: PricingStructure) -> BoundReferenceResolver:
    return BoundReferenceResolver(pricing_structure, ps_to_tariff_resolver, None)


def transformer_end(ratio_tap_changer: RatioTapChanger) -> BoundReferenceResolver:
    return BoundReferenceResolver(ratio_tap_changer, rtc_to_te_resolver, te_to_rtc_resolver)


def control(remote_control: RemoteControl) -> BoundReferenceResolver:
    return BoundReferenceResolver(remote_control, rc_to_cont_resolver, control_to_remote_control_resolver)


def measurement(remote_source: RemoteSource) -> BoundReferenceResolver:
    return BoundReferenceResolver(remote_source, rs_to_meas_resolver, meas_to_rs_resolver)


def geographical_region(sub_geographical_region: SubGeographicalRegion) -> BoundReferenceResolver:
    return BoundReferenceResolver(sub_geographical_region, sgr_to_gr_resolver, gr_to_sgr_resolver)


def substations(sub_geographical_region: SubGeographicalRegion) -> BoundReferenceResolver:
    return BoundReferenceResolver(sub_geographical_region, sgr_to_sub_resolver, sub_to_sgr_resolver)


def normal_energizing_feeders(substation: Substation) -> BoundReferenceResolver:
    return BoundReferenceResolver(substation, sub_to_feeder_resolver, feeder_to_nes_resolver)


def sub_geographical_region(substation: Substation) -> BoundReferenceResolver:
    return BoundReferenceResolver(substation, sub_to_sgr_resolver, sgr_to_sub_resolver)


def circuits(substation: Substation) -> BoundReferenceResolver:
    return BoundReferenceResolver(substation, sub_to_circuit_resolver, circuit_to_sub_resolver)


def normal_energized_loops(substation: Substation) -> BoundReferenceResolver:
    return BoundReferenceResolver(substation, sub_to_eloop_resolver, loop_to_esub_resolver)


def loops(substation: Substation) -> BoundReferenceResolver:
    return BoundReferenceResolver(substation, sub_to_loop_resolver, loop_to_sub_resolver)


def conducting_equipment(terminal: Terminal) -> BoundReferenceResolver:
    return BoundReferenceResolver(terminal, term_to_ce_resolver, cond_equip_to_terminal_resolver)


def connectivity_node(terminal: Terminal) -> BoundReferenceResolver:
    return BoundReferenceResolver(terminal, term_to_cn_resolver, conn_node_to_term_resolver)


def te_base_voltage(transformer_end: TransformerEnd) -> BoundReferenceResolver:
    return BoundReferenceResolver(transformer_end, te_to_bv_resolver, None)


def ratio_tap_changer(transformer_end: TransformerEnd) -> BoundReferenceResolver:
    return BoundReferenceResolver(transformer_end, te_to_rtc_resolver, rtc_to_te_resolver)


def te_terminal(transformer_end: TransformerEnd) -> BoundReferenceResolver:
    return BoundReferenceResolver(transformer_end, te_to_term_resolver, None)


def end_devices(usage_point: UsagePoint) -> BoundReferenceResolver:
    return BoundReferenceResolver(usage_point, up_to_ed_resolver, ed_to_up_resolver)


def up_equipment(usage_point: UsagePoint) -> BoundReferenceResolver:
    return BoundReferenceResolver(usage_point, up_to_eq_resolver, eq_to_up_resolver)


def usage_point_location(usage_point: UsagePoint) -> BoundReferenceResolver:
    return BoundReferenceResolver(usage_point, up_to_loc_resolver, None)


def loop(circuit: Circuit) -> BoundReferenceResolver:
    return BoundReferenceResolver(circuit, circuit_to_loop_resolver, loop_to_circuit_resolver)


def end_terminal(circuit: Circuit) -> BoundReferenceResolver:
    return BoundReferenceResolver(circuit, circuit_to_term_resolver, None)


def end_substation(circuit: Circuit) -> BoundReferenceResolver:
    return BoundReferenceResolver(circuit, circuit_to_sub_resolver, sub_to_circuit_resolver)


def loop_circuits(loop: Loop) -> BoundReferenceResolver:
    return BoundReferenceResolver(loop, loop_to_circuit_resolver, circuit_to_loop_resolver)


def loop_substations(loop: Loop) -> BoundReferenceResolver:
    return BoundReferenceResolver(loop, loop_to_sub_resolver, sub_to_loop_resolver)


def loop_energizing_substations(loop: Loop) -> BoundReferenceResolver:
    return BoundReferenceResolver(loop, loop_to_esub_resolver, sub_to_eloop_resolver)



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

from typing import Optional, Generator, List

from zepben.evolve.model.cim.iec61970.base.core.identified_object import IdentifiedObject
from zepben.evolve.util import nlen, get_by_mrid, ngen, safe_remove

__all__ = ["GeographicalRegion", "SubGeographicalRegion"]


class GeographicalRegion(IdentifiedObject):
    """
    A geographical region of a power system network phases.
    """
    _sub_geographical_regions: Optional[List[SubGeographicalRegion]] = None

    def __init__(self, sub_geographical_regions: List[SubGeographicalRegion] = None):
        if sub_geographical_regions:
            for sgr in sub_geographical_regions:
                self.add_sub_geographical_region(sgr)

    def num_sub_geographical_regions(self) -> int:
        """
        Returns The number of `SubGeographicalRegion`s associated with this `GeographicalRegion`
        """
        return nlen(self._sub_geographical_regions)

    @property
    def sub_geographical_regions(self) -> Generator[SubGeographicalRegion, None, None]:
        """
        The `SubGeographicalRegion`s of this `GeographicalRegion`.
        """
        return ngen(self._sub_geographical_regions)

    def get_sub_geographical_region(self, mrid: str) -> SubGeographicalRegion:
        """
        Get the `SubGeographicalRegion` for this `GeographicalRegion` identified by `mrid`

        `mrid` The mRID of the required `SubGeographicalRegion`
        Returns The `SubGeographicalRegion` with the specified `mrid` if it exists
        Raises `KeyError` if `mrid` wasn't present.
        """
        return get_by_mrid(self._sub_geographical_regions, mrid)

    def add_sub_geographical_region(self, sub_geographical_region: SubGeographicalRegion) -> GeographicalRegion:
        """
        Associate a `SubGeographicalRegion` with this `GeographicalRegion`

        `sub_geographical_region` The `SubGeographicalRegion` to associate with this `GeographicalRegion`.
        Returns A reference to this `GeographicalRegion` to allow fluent use.
        Raises `ValueError` if another `SubGeographicalRegion` with the same `mrid` already exists for this `GeographicalRegion`.
        """
        if self._validate_reference(sub_geographical_region, self.get_sub_geographical_region, "A SubgeographicalRegion"):
            return self
        self._sub_geographical_regions = list() if self._sub_geographical_regions is None else self._sub_geographical_regions
        self._sub_geographical_regions.append(sub_geographical_region)
        return self

    def remove_sub_geographical_region(self, sub_geographical_region: SubGeographicalRegion) -> GeographicalRegion:
        """
        Disassociate `sub_geographical_region` from this `GeographicalRegion`
        `sub_geographical_region` The `SubGeographicalRegion` to disassociate from this `GeographicalRegion`.
        Returns A reference to this `GeographicalRegion` to allow fluent use.
        Raises `ValueError` if `sub_geographical_region` was not associated with this `GeographicalRegion`.
        """
        self._sub_geographical_regions = safe_remove(self._sub_geographical_regions, sub_geographical_region)
        return self

    def clear_sub_geographical_regions(self) -> GeographicalRegion:
        """
        Clear all SubGeographicalRegions.
        Returns A reference to this `GeographicalRegion` to allow fluent use.
        """
        self._sub_geographical_regions = None
        return self


class SubGeographicalRegion(IdentifiedObject):
    """
    A subset of a geographical region of a power system network model.
    """

    geographical_region: Optional[GeographicalRegion] = None
    """The geographical region to which this sub-geographical region is within."""

    _substations: Optional[List[Substation]] = None

    def __init__(self, substations: List[Substation] = None):
        if substations:
            for sub in substations:
                self.add_substation(sub)

    def num_substations(self) -> int:
        """
        Returns The number of `zepben.evolve.iec61970.base.core.substation.Substation`s associated with this `SubGeographicalRegion`
        """
        return nlen(self._substations)

    @property
    def substations(self) -> Generator[Substation, None, None]:
        """
        All substations belonging to this sub geographical region.
        """
        return ngen(self._substations)

    def get_substation(self, mrid: str) -> Substation:
        """
        Get the `zepben.evolve.iec61970.base.core.substation.Substation` for this `SubGeographicalRegion` identified by `mrid`

        `mrid` the mRID of the required `zepben.evolve.iec61970.base.core.substation.Substation`
        Returns The `zepben.evolve.iec61970.base.core.substation.Substation` with the specified `mrid` if it exists
        Raises `KeyError` if `mrid` wasn't present.
        """
        return get_by_mrid(self._substations, mrid)

    def add_substation(self, substation: Substation) -> SubGeographicalRegion:
        """
        Associate a `Substation` with this `GeographicalRegion`

        `substation` the `zepben.evolve.iec61970.base.core.substation.Substation` to associate with this `SubGeographicalRegion`.
        Returns A reference to this `SubGeographicalRegion` to allow fluent use.
        Raises `ValueError` if another `zepben.evolve.iec61970.base.core.substation.Substation` with the same `mrid` already exists for this `GeographicalRegion`.
        """
        if self._validate_reference(substation, self.get_substation, "A Substation"):
            return self
        self._substations = list() if self._substations is None else self._substations
        self._substations.append(substation)
        return self

    def remove_substation(self, substation: Substation) -> SubGeographicalRegion:
        """
        Disassociate `substation` from this `GeographicalRegion`

        `substation` The `zepben.evolve.iec61970.base.core.substation.Substation` to disassociate from this `SubGeographicalRegion`.
        Returns A reference to this `SubGeographicalRegion` to allow fluent use.
        Raises `ValueError` if `substation` was not associated with this `SubGeographicalRegion`.
        """
        self._substations = safe_remove(self._substations, substation)
        return self

    def clear_substations(self) -> SubGeographicalRegion:
        """
        Clear all `Substations`.
        Returns A reference to this `SubGeographicalRegion` to allow fluent use.
        """
        self._substations = None
        return self



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
from unittest.mock import MagicMock

import pytest
from zepben.evolve import CableInfo, NetworkService
from zepben.evolve.streaming.put.producer import NetworkProducerClient, ProducerClient, SyncNetworkProducerClient
from zepben.protobuf.np.np_pb2_grpc import NetworkProducerStub


class TestProducer(object):

    def test_send(self):
        stub: NetworkProducerStub = MagicMock(["CreateNetwork", "CompleteNetwork", "CreateCableInfo"])
        producer_client: SyncNetworkProducerClient = SyncNetworkProducerClient(stub=stub)
        service: NetworkService = NetworkService()

        cable_info = CableInfo("id1")
        service.add(cable_info)
        producer_client.send(service)
        stub.CreateNetwork.assert_called_once()
        stub.CreateCableInfo.assert_called_once()
        stub.CompleteNetwork.assert_called_once()

    @pytest.mark.asyncio
    async def test_send_all_services(self):
        pass
        # network_stub = MagicMock(["CreateNetwork", "CompleteNetwork", "CreateCableInfo"])
        # diagram_stub = MagicMock(["CreateDiagramService", "CompleteDiagramService", "CreateDiagram"])
        # customer_stub = MagicMock(["CreateCustomerService", "CompleteCustomerService", "CreateCustomer"])
        #
        # producer_client = ProducerClient(stub=stub)
        # service: NetworkService = NetworkService()



#  Copyright 2020 Zeppelin Bend Pty Ltd
# 
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations
from abc import abstractmethod

from dataclassy import dataclass


from zepben.evolve.services.network.tracing.traversals.queue import FifoQueue, LifoQueue, PriorityQueue, Queue
from zepben.evolve.exceptions import TracingException
from zepben.evolve.services.network.tracing.traversals import Tracker
from typing import List, Callable, Awaitable, TypeVar, Generic, Set, Iterable
from enum import Enum

__all__ = ["SearchType", "create_queue", "BaseTraversal", "Traversal"]
T = TypeVar('T')


class SearchType(Enum):
    BREADTH = 1
    DEPTH = 2
    PRIORITY = 3


def create_queue(search_type):
    if search_type == SearchType.DEPTH:
        return LifoQueue()
    elif search_type == SearchType.BREADTH:
        return FifoQueue()
    elif search_type == SearchType.PRIORITY:
        return PriorityQueue()


@dataclass(slots=True)
class BaseTraversal(Generic[T]):
    """
    A basic traversal implementation that can be used to traverse any type of item.
    This class is asyncio compatible. Stop condition and step action callbacks are called with await.

    A stop condition is a callback function that must return a boolean indicating whether the Tracer should stop
    processing the current branch. Tracing will only stop when either:
        - All branches have been exhausted, or
        - A stop condition has returned true on every possible branch.
    Stop conditions will be called prior to applying any callbacks, but the stop will only occur after all actions
    have been applied.

    Step actions are functions to be called on each item visited in the trace. These are called after the stop conditions are evaluated, and each action is
    passed the current `zepben.evolve.network.tracing.connectivity.ConnectivityResult` as well as the `stopping` state (True if the trace is stopping after
    the current `ConnectivityResult, False otherwise). Thus, the signature of each step action must be:
    :func: action(cr: `zepben.evolve.tracing.ConnectivityResult`, is_stopping: bool) -> None
    """
    start_item: T = None
    """The starting item for this `BaseTraversal`"""

    stop_conditions: List[Callable[[T], Awaitable[bool]]] = []
    """A list of callback functions, to be called in order with the current item."""

    step_actions: List[Callable[[T, bool], Awaitable[None]]] = []
    """A list of callback functions, to be called on each item."""

    _has_run: bool = False
    """Whether this traversal has run """

    _running: bool = False
    """Whether this traversal is currently running"""

    async def matches_stop_condition(self, item: T):
        """
        Checks all the stop conditions for the passed in item and returns true if any match.
        This calls all registered stop conditions even if one has already returned true to make sure everything is
        notified about this item.
        Each stop condition will be awaited and thus must be an async function.

        `item` The item to pass to the stop conditions.
        Returns True if any of the stop conditions return True.
        """
        stop = False
        for cond in self.stop_conditions:
            stop = stop or await cond(item)
        return stop

    def add_stop_condition(self, cond: Callable[[T], Awaitable[bool]]):
        """
        Add a callback to check whether the current item in the traversal is a stop point.
        If any of the registered stop conditions return true, the traversal will not call the callback to queue more items.
        Note that a match on a stop condition doesn't necessarily stop the traversal, it just stops traversal of the current branch.
                                                                                                                      
        `cond` A function that if returns true will cause the traversal to stop traversing the branch.
        Returns this traversal instance.
        """
        self.stop_conditions.append(cond)

    def add_step_action(self, action: Callable[[T, bool], Awaitable[None]]) -> BaseTraversal[T]:
        """
        Add a callback which is called for every item in the traversal (including the starting item).
                                                                                                             
        `action` Action to be called on each item in the traversal, passing if the trace will stop on this step.
        Returns this traversal instance.
        """
        self.step_actions.append(action)
        return self

    def copy_stop_conditions(self, other: BaseTraversal[T]):
        """Copy the stop conditions from `other` to this `BaseTraversal`."""
        self.stop_conditions.extend(other.stop_conditions)

    def copy_step_actions(self, other: BaseTraversal[T]):
        """Copy the step actions from `other` to this `BaseTraversal`."""
        self.step_actions.extend(other.step_actions)

    def clear_stop_conditions(self):
        """Clear all stop conditions."""
        self.stop_conditions.clear()

    def clear_step_actions(self):
        """Clear all step actions"""
        self.step_actions.clear()

    async def apply_step_actions(self, item: T, is_stopping: bool):
        """
        Calls all the step actions with the passed in item.
        Each action will be awaited.
        `item` The item to pass to the step actions.
        `is_stopping` Indicates if the trace will stop on this step.
        """
        for action in self.step_actions:
            await action(item, is_stopping)

    def _reset_run_flag(self):
        if self._running:
            raise TracingException("Can't reset when Traversal is currently executing.")
        self._has_run = False

    @abstractmethod
    def reset(self):
        """
        Reset this traversal. Should take care to reset all fields and queues so that the traversal can be reused.
        """
        raise NotImplementedError()

    async def trace(self, start_item: T = None, can_stop_on_start_item: bool = True):
        """
        Perform a trace across the network from `start_item`, applying actions to each piece of equipment encountered
        until all branches of the network are exhausted, or a stop condition succeeds and we cannot continue any further.
        When a stop condition is reached, we will stop tracing that branch of the network and continue with other branches.
        `start_item` The starting point. Must implement :func:`zepben.evolve.ConductingEquipment::get_connectivity`
                           which allows tracing over the terminals in a network.
        `can_stop_on_start_item` If it's possible for stop conditions to apply to the start_item.
        """
        if self._running:
            raise TracingException("Traversal is already running.")

        if self._has_run:
            raise TracingException("Traversal must be reset before reuse.")

        self._running = True
        self._has_run = True
        self.start_item = start_item if start_item is not None else self.start_item
        await self._run_trace(can_stop_on_start_item)
        self._running = False

    @abstractmethod
    async def _run_trace(self, can_stop_on_start_item: bool = True):
        """
        Extend and implement your tracing algorithm here.
        `start_item` The starting object to commence tracing. Must implement :func:`zepben.evolve.ConductingEquipment.get_connectivity`
        `can_stop_on_start_item` Whether to
        """
        raise NotImplementedError()


class Traversal(BaseTraversal):
    """
    A basic traversal implementation that can be used to traverse any type of item.

    The traversal gets the next items to be traversed to by calling a user provided callback (next_), with the current
    item of the traversal. This function should return a list of ConnectivityResult's, that will get added to the
    process_queue for processing.

    Different `SearchType`'s types can be used to provide different trace types via the `process_queue`.
    The default `Depth` will utilise a `LifoQueue` to provide a depth-first search of the network, while a `Breadth`
    will use a FIFO `Queue` breadth-first search. More complex searches can be achieved with `Priority`, which
    will use a PriorityQueue under the hood.

    The traversal also requires a `zepben.evolve.traversals.tracker.Tracker` to be supplied. This gives flexibility
    to track items in unique ways, more than just "has this item been visited" e.g. visiting more than once,
    visiting under different conditions etc.
    """
    queue_next: Callable[[T, Set[T]], Iterable[T]]
    """A function that will return a list of `T` to add to the queue. The function must take the item to queue and optionally a set of already visited items."""

    process_queue: Queue
    """Dictates the type of search to be performed on the network graph. Breadth-first, Depth-first, and Priority based searches are possible."""

    tracker: Tracker = Tracker()
    """A `zepben.evolve.traversals.tracker.Tracker` for tracking which items have been seen. If not provided a `Tracker` will be created for this trace."""

    async def _run_trace(self, can_stop_on_start_item: bool = True):
        """
        Run's the trace. Stop conditions and step_actions are called with await, so you can utilise asyncio when
        performing a trace if your step actions or conditions are IO intensive. Stop conditions and
        step actions will always be called for each item in the order provided.
        `can_stop_on_start_item` Whether the trace can stop on the start_item. Actions will still be applied to
                                       the start_item.
        """
        if self.start_item is None:
            try:
                self.start_item = self.process_queue.get()
            except IndexError:
                raise TracingException("Starting item wasn't specified and the process queue is empty. Cannot start the trace.")

        self.tracker.visit(self.start_item)
        # If we can't stop on the start item we don't run any stop conditions. if this causes a problem for you,
        # you should run the stop conditions for the start item prior to running the traversal.
        stopping = can_stop_on_start_item and await self.matches_stop_condition(self.start_item)
        await self.apply_step_actions(self.start_item, stopping)
        if not stopping:
            for x in self.queue_next(self.start_item, self.tracker.visited):
                self.process_queue.put(x)

        while not self.process_queue.empty():
            current = self.process_queue.get()
            if self.tracker.visit(current):
                stopping = await self.matches_stop_condition(current)
                await self.apply_step_actions(current, stopping)
                if not stopping:
                    for x in self.queue_next(current, self.tracker.visited):
                        self.process_queue.put(x)

    def reset(self):
        self._reset_run_flag()
        self.process_queue.queue.clear()
        self.tracker.clear()


def _depth_trace(start_item, stop_on_start_item=True, stop_fn=None, equip_fn=None, term_fn=None):
    equips_to_trace = []
    traced = set()
    for t in start_item.terminals:
        traced.add(t.mrid)
    if stop_on_start_item:
        yield start_item
    equips_to_trace.append(start_item)
    while equips_to_trace:
        try:
            equip = equips_to_trace.pop()
        except IndexError:  # No more equipment
            break
        # Explore all connectivity nodes for this equipments terminals,
        # and set upstream on each terminal.
        for terminal in equip.terminals:
            conn_node = terminal.connectivity_node
            for term in conn_node:
                # keep this
                if term.mrid in traced:
                    continue
                if term != terminal:
                    if not term.conducting_equipment.connected():
                        continue
                    equips_to_trace.append(term.conducting_equipment)
                    yield term.conducting_equipment
                # Don't trace over a terminal twice to stop us from reversing direction
                traced.add(term.mrid)



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
import logging
from typing import Optional, Callable, Set, Iterable, TypeVar

from zepben.evolve.model.cim.iec61970.base.core.conducting_equipment import ConductingEquipment
from zepben.evolve.model.cim.iec61970.base.core.terminal import Terminal
from zepben.evolve.model import SinglePhaseKind
from zepben.evolve.model.phasedirection import PhaseDirection
from zepben.evolve.services.network.tracing.phase_step import PhaseStep
from zepben.evolve.services.network.tracing.phase_status import PhaseStatus, current_phases, normal_phases
from zepben.evolve.services.network.tracing.connectivity import get_connected_equipment, get_connectivity
from zepben.evolve.services.network.tracing.util import currently_open, normally_open
from zepben.evolve.services.network.tracing.traversals import Traversal, depth_first, Queue, PriorityQueue

__all__ = ["queue_next_terminal", "normal_downstream_trace", "create_basic_depth_trace", "connected_equipment_trace", "current_downstream_trace"]

tracing_logger = logging.getLogger("queue_next")

T = TypeVar("T")


def connected_equipment_trace():
    return create_basic_depth_trace(conducting_equipment_queue_next)


def create_basic_depth_trace(queue_next: Callable[[T, Set[T]], Iterable[T]]):
    return Traversal(queue_next, depth_first())


def conducting_equipment_queue_next(conducting_equipment: Optional[ConductingEquipment], exclude: Optional[Set] = None) -> Iterable[ConductingEquipment]:
    """
    Get the next `ConductingEquipment` to queue next as determined by `conducting_equipment`s connectivity.
    `conducting_equipment` the `ConductingEquipment` to fetch connected equipment for.
    `exclude` Any `ConductingEquipment` that should be excluded from the result.
    Returns a list of `ConductingEquipment` that should be queued next.
    """
    if exclude is None:
        exclude = []

    if conducting_equipment:
        crs = get_connected_equipment(conducting_equipment, exclude)
        return [cr.to_equip for cr in crs if cr.to_equip and cr.to_equip not in exclude]


def current_downstream_trace(queue: Queue = None, **kwargs):
    """
    Create a downstream trace over current phases
    `queue` Queue to use for this trace. Defaults to a `zepben.evolve.traversals.queue.PriorityQueue`
    `kwargs` Args to be passed to `zepben.evolve.Traversal`
    Returns A `zepben.evolve.traversals.Traversal`
    """
    return Traversal(queue_next=_create_downstream_queue_next(currently_open, current_phases), process_queue=queue, **kwargs)


def _create_downstream_queue_next(open_test: Callable[[ConductingEquipment, Optional[SinglePhaseKind]], bool],
                                  active_phases: Callable[[Terminal, SinglePhaseKind], PhaseStatus]):
    """
    Creates a queue_next function from the given open test and phase selector for use with traversals.
    `open_test` Function that takes a ConductingEquipment and a phase and returns whether the phase on the equipment is open (True) or closed (False).
    `active_phases` A `zepben.evolve.phase_status.PhaseStatus`
    Returns A queue_next function for use with `zepben.evolve.BaseTraversal` classes
    """
    def qn(phase_step, visited):
        connected_terms = []
        if not phase_step:
            return connected_terms
        out_phases = set()
        for term in phase_step.conducting_equipment.terminals:
            _get_phases_with_direction(open_test, active_phases, term, phase_step.phases, PhaseDirection.OUT, out_phases)

            if out_phases:
                crs = get_connectivity(term, out_phases)
                for cr in crs:
                    if cr.to_equip is not None:
                        if cr.to_equip in visited:
                            continue
                        connected_terms.append(PhaseStep(cr.to_equip, out_phases, cr.from_equip))
        return connected_terms
    return qn


def queue_next_terminal(item, exclude: Optional[Set] = None):
    """
    Wrapper tracing queue function for fetching the terminals that should be queued based on their connectivity

    `item` The Terminal to fetch connected `zepben.evolve.iec61970.base.core.terminal.Terminal`s for.
    `exclude` set of `Terminal`s to be excluded from queuing.
    Returns a list of `zepben.evolve.iec61970.base.core.terminal.Terminal`s to be queued
    """
    other_terms = item.get_other_terminals()
    if not other_terms:
        # If there are no other terminals we get connectivity for this one and return that. Note that this will
        # also return connections for EnergyConsumer's, but upstream will be covered by the exclude parameter and thus
        # should yield an empty list.
        to_terms = [cr.to_terminal for cr in item.get_connectivity(exclude=exclude)]
        if len(to_terms) > 0:
            tracing_logger.debug(f"Queuing {to_terms[0].mrid} from single terminal equipment {item.mrid}")
        return to_terms

    crs = []
    for term in other_terms:
        crs.extend(term.get_connectivity(exclude=exclude))

    to_terms = [cr.to_terminal for cr in crs]
    tracing_logger.debug(f"Queuing terminals: [{', '.join(t.mrid for t in to_terms)}] from {item.mrid}")
    return to_terms


def normal_downstream_trace(queue: Queue = None, **kwargs):
    """
    Create a downstream trace over nominal phases.

    `queue` Queue to use for this trace. Defaults to a `zepben.evolve.traversals.queue.PriorityQueue`
    `kwargs` Args to be passed to `zepben.evolve.Traversal`
    Returns A `zepben.evolve.traversals.Traversal`
    """
    if queue is None:
        queue = PriorityQueue()
    return Traversal(queue_next=_create_downstream_queue_next(normally_open, normal_phases), process_queue=queue, **kwargs)


def _get_phases_with_direction(open_test: Callable[[ConductingEquipment, Optional[SinglePhaseKind]], bool],
                               active_phases: Callable[[Terminal, SinglePhaseKind], PhaseStatus],
                               terminal: Terminal,
                               candidate_phases: Set[SinglePhaseKind],
                               direction: PhaseDirection,
                               matched_phases: Set[SinglePhaseKind]):
    """
    Adds the closed phases from `terminal` in a specified `zepben.evolve.model.phasedirection.PhaseDirection` to `matched_phases`.

    `open_test` Function that takes a ConductingEquipment and a phase and returns whether the phase on the equipment is open (True) or closed (False).
    `active_phases` A `zepben.evolve.phase_status.PhaseStatus`
    `terminal` `zepben.evolve.cim.iec61970.base.core.terminal.Terminal` to retrieve phases for
    `filter_cores` The phases of `terminal` to test.
    `direction` The `zepben.evolve.model.phasedirection.PhaseDirection` to check against.
    `matched_phases` The set of matched phases to add to.
    """
    if terminal.conducting_equipment is None:
        raise TraceException(f"Terminal {terminal} did not have an associated ConductingEquipment, cannot get phases.")
    for phase in candidate_phases:
        if phase in terminal.phases.single_phases and not open_test(terminal.conducting_equipment, phase):
            if active_phases(terminal, phase).direction().has(direction):
                matched_phases.add(phase)


class TraceException(Exception):
    pass





#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

from typing import Optional, TypeVar, Generic, Callable, List, Union

from dataclassy import dataclass

__all__ = ["GrpcResult", "GrpcClient"]

T = TypeVar("T")


@dataclass(slots=True)
class GrpcResult(Generic[T]):
    result: Optional[Union[T, Exception]]
    was_error_handled: bool = False

    @property
    def was_successful(self):
        return not self.was_failure

    @property
    def was_failure(self):
        return isinstance(self.result, Exception)

    def on_success(self, handler: Callable[[T], None]) -> GrpcResult[T]:
        """Calls `handler` with the `result` if this `was_successful`"""
        if self.was_successful:
            handler(self.result)
        return self

    def on_error(self, handler: Callable[[Exception, bool], GrpcResult[T]]) -> GrpcResult[T]:
        """Calls `handler` with the `thrown` exception and `self.was_error_handled` if `self.was_failure`."""
        if self.was_failure and self.was_error_handled:
            return handler(self.result, self.was_error_handled)
        return self

    def on_handled_error(self, handler: Callable[[Exception], None]) -> GrpcResult[T]:
        """Calls `handler` with the `thrown` exception if `self.was_failure` only if `self.was_error_handled`."""
        if self.was_failure and self.was_error_handled:
            handler(self.result)
        return self

    def on_unhandled_error(self, handler: Callable[[Exception], None]) -> GrpcResult[T]:
        """Calls `handler` with the `thrown` exception if `self.was_failure` only if not `self.was_error_handled`."""
        if self.was_failure and not self.was_error_handled:
            handler(self.result)
        return self

    def throw_on_error(self) -> GrpcResult[T]:
        """Throws `self.result` if `self.was_failure`"""
        if self.was_failure:
            raise self.result
        return self

    def throw_on_unhandled_error(self) -> GrpcResult[T]:
        """Throws `self.result` only if `self.was_failure` and not `self.was_error_handled`"""
        if self.was_failure and not self.was_error_handled:
            raise self.result
        return self


@dataclass(init=False, slots=True)
class GrpcClient(object):
    error_handlers: List[Callable[[Exception], bool]] = []

    def __init__(self, error_handlers: List[Callable[[Exception], bool]] = None):
        if error_handlers:
            self.error_handlers = error_handlers.copy()
        else:
            self.error_handlers = list()

    def try_handle_error(self, e: Exception) -> bool:
        for handler in self.error_handlers:
            if handler(e):
                return True
        return False

    async def try_rpc(self, rpc: Callable[[], IdentifiedObject]) -> GrpcResult:
        try:
            return GrpcResult(await rpc())
        except Exception as e:
            return GrpcResult(e, self.try_handle_error(e))



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations
import logging
from enum import Enum
from typing import Dict, List

from zepben.evolve.model import Measurement
from zepben.evolve.services.common import BaseService
from zepben.evolve.model import ConnectivityNode
from zepben.evolve.model import EnergySource
from zepben.evolve.services.network.tracing import SetPhases
from pathlib import Path

__all__ = ["connect", "NetworkService"]
logger = logging.getLogger(__name__)
TRACED_NETWORK_FILE = str(Path.home().joinpath(Path("traced.json")))


class ProcessStatus(Enum):
    PROCESSED = 0
    INVALID = 1
    SKIPPED = 2


def connect(terminal: Terminal, connectivity_node: ConnectivityNode):
    """
    Connect a `zepben.evolve.iec61970.base.core.terminal.Terminal`` to a `ConnectivityNode`
    `terminal` The `zepben.evolve.iec61970.base.core.terminal.Terminal` to connect.
    `connectivity_node` The `ConnectivityNode` to connect ``zepben.evolve.iec61970.base.core.terminal.Terminal` to.
    """
    terminal.connect(connectivity_node)
    connectivity_node.add_terminal(terminal)


def _attempt_to_reuse_connection(terminal1: Terminal, terminal2: Terminal) -> ProcessStatus:
    """
    Attempt to connect two `zepben.evolve.iec61970.base.core.terminal.Terminal`s.
    Returns `ProcessStatus` reflecting whether the connection was reused. PROCESSED if a connection was
    established, INVALID if it couldn't be, and SKIPPED if neither terminal had an existing `ConnectivityNode`.
    """
    cn1 = terminal1.connectivity_node
    cn2 = terminal2.connectivity_node

    if cn1 is not None:
        if cn2 is not None:
            if cn1 is cn2:
                return ProcessStatus.PROCESSED
            elif connect(terminal2, cn1.mrid):
                return ProcessStatus.PROCESSED
            return ProcessStatus.INVALID
    elif cn2 is not None:
        return ProcessStatus.PROCESSED if connect(terminal1, cn2.mrid) else ProcessStatus.INVALID
    return ProcessStatus.SKIPPED


class NetworkService(BaseService):
    """
    A full representation of the power network.
    Contains a map of equipment (string ID's -> Equipment/Nodes/etc)
    **All** `IdentifiedObject's` submitted to this Network **MUST** have unique mRID's!

    Attributes -
        metrics_store : Storage for meter measurement data associated with this network.
    """

    name: str = "network"
    _connectivity_nodes: Dict[str, ConnectivityNode] = dict()
    _auto_cn_index: int = 0
    _measurements: Dict[str, List[Measurement]] = []

    def __init__(self):
        self._objectsByType[ConnectivityNode] = self._connectivity_nodes

    def get_measurements(self, mrid: str, t: type) -> List[Measurement]:
        """
        Get all measurements of type `t` associated with the given `mrid`.
                                                                                                              
        The `mrid` should be either a `zepben.evolve.iec61970.base.core.power_system_resource.PowerSystemResource` or a
        `zepben.evolve.iec61970.base.core.terminal.Terminal` MRID that is assigned to the corresponding fields on the measurements.
        Returns all `Measurement`s indexed by `mrid` in this service.
        Raises `KeyError` if `mrid` isn't present in this service.
        """
        return [meas for meas in self._measurements[mrid] if isinstance(meas, t)]

    def add_measurement(self, measurement: Measurement) -> bool:
        """
        Add a `zepben.evolve.cim.iec61970.base.meas.measurement.Measurement` to this `NetworkService`

        `measurement` The `Measurement` to add.
        Returns `True` if `measurement` was added, `False` otherwise
        """
        return self._index_measurement(measurement) and self.add(measurement)

    def remove_measurement(self, measurement) -> bool:
        """
        Remove a `zepben.evolve.cim.iec61970.base.meas.measurement.Measurement` from this `NetworkService`

        `measurement` The `Measurement` to remove.
        Returns `True` if `measurement` was removed, `False` otherwise
        """
        self._remove_measurement_index(measurement)
        return self.remove(measurement)

    def connect_by_mrid(self, terminal: Terminal, connectivity_node_mrid: str) -> bool:
        """
        Connect a `zepben.evolve.iec61970.base.core.terminal.Terminal` to the `ConnectivityNode` with mRID `connectivity_node_mrid`
        `terminal` The `zepben.evolve.iec61970.base.core.terminal.Terminal` to connect.
        `connectivity_node_mrid` The mRID of the `ConnectivityNode`. Will be created in the `Network` if it
        doesn't already exist.
        Returns True if the connection was made or already existed, False if `zepben.evolve.iec61970.base.core.terminal.Terminal` was already connected to a
        different `ConnectivityNode`
        """
        if not connectivity_node_mrid:
            return False

        if terminal.connectivity_node:
            return connectivity_node_mrid == terminal.connectivity_node.mrid

        cn = self.add_connectivitynode(connectivity_node_mrid)
        connect(terminal, cn)
        return True

    def connect_terminals(self, terminal1: Terminal, terminal2: Terminal) -> bool:
        """
        Connect two `zepben.evolve.iec61970.base.core.terminal.Terminal`s
        Returns True if the `zepben.evolve.iec61970.base.core.terminal.Terminal`s could be connected, False otherwise.
        """
        status = _attempt_to_reuse_connection(terminal1, terminal2)
        if status == ProcessStatus.PROCESSED:
            return True
        elif status == ProcessStatus.INVALID:
            return False

        cn = self.add_connectivitynode(self._generate_cn_mrid())
        connect(terminal2, cn)
        connect(terminal1, cn)

        return True

    def _generate_cn_mrid(self):
        mrid = f"generated_cn_{self._auto_cn_index}"
        while mrid in self._connectivity_nodes:
            self._auto_cn_index += 1
            mrid = f"generated_cn_{self._auto_cn_index}"
        return mrid

    def disconnect(self, terminal: Terminal):
        """
        Disconnect a `zepben.evolve.iec61970.base.core.terminal.Terminal`` from its `ConnectivityNode`. Will also remove the `ConnectivityNode` from this
        `Network` if it no longer has any terminals.
        `terminal` The `zepben.evolve.iec61970.base.core.terminal.Terminal` to disconnect.
        """
        cn = terminal.connectivity_node
        if cn is None:
            return
        cn.remove_terminal(terminal)
        terminal.disconnect()
        if cn.num_terminals() == 0:
            del self._connectivity_nodes[cn.mrid]

    def disconnect_by_mrid(self, connectivity_node_mrid: str):
        """
        Disconnect a `ConnectivityNode` from this `Network`. Will disconnect all ``zepben.evolve.iec61970.base.core.terminal.Terminal`s from the
        `ConnectivityNode`
        `connectivity_node_mrid` The mRID of the `ConnectivityNode` to disconnect.
        Raises `KeyError` if there is no `ConnectivityNode` for `connectivity_node_mrid`
        """
        cn = self._connectivity_nodes[connectivity_node_mrid]
        if cn is not None:
            for term in cn.terminals:
                term.disconnect()
            cn.clear_terminals()
            del self._connectivity_nodes[connectivity_node_mrid]

    def get_primary_sources(self):
        """
        Get the primary source for this network. All directions are applied relative to this EnergySource
        Returns The primary EnergySource
        """
        return [source for source in self._objectsByType[EnergySource].values() if source.has_phases()]

    def add_connectivitynode(self, mrid: str):
        """
        Add a connectivity node to the network.
        `mrid` mRID of the ConnectivityNode
        Returns A new ConnectivityNode with `mrid` if it doesn't already exist, otherwise the existing
                 ConnectivityNode represented by `mrid`
        """
        if mrid not in self._connectivity_nodes:
            self._connectivity_nodes[mrid] = ConnectivityNode(mrid=mrid)
            return self._connectivity_nodes[mrid]
        else:
            return self._connectivity_nodes[mrid]

    async def set_phases(self):
        set_phases = SetPhases()
        await set_phases.run(self)

    def _index_measurement(self, measurement: Measurement, mrid: str) -> bool:
        if not mrid:
            return False

        if mrid in self._measurements:
            for meas in self._measurements[mrid]:
                if meas.mrid == measurement.mrid:
                    return False
            else:
                self._measurements[mrid].append(measurement)
                return True
        else:
            self._measurements[mrid] = [measurement]
            return True

    def _remove_measurement_index(self, measurement: Measurement):
        try:
            self._measurements[measurement.terminal_mrid].remove(measurement)
        except KeyError:
            pass
        try:
            self._measurements[measurement.power_system_resource_mrid].remove(measurement)
        except KeyError:
            pass







#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

from zepben.evolve.model.cim.iec61970.base.core.conducting_equipment import ConductingEquipment

__all__ = ["Connector", "Junction"]


class Connector(ConductingEquipment):
    """
    A conductor, or group of conductors, with negligible impedance, that serve to connect other conducting equipment
    within a single substation and are modelled with a single logical terminal.
    """
    pass


class Junction(Connector):
    """
    A point where one or more conducting equipments are connected with zero resistance.
    """
    pass




#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from collections import namedtuple
from zepben.evolve.services.network.tracing.traversals.tracing import Traversal
from zepben.evolve.services.network.tracing import queue_next_equipment
from zepben.evolve.services.network.tracing.traversals.queue import LifoQueue

__all__ = ["extract_latlongs"]

LatLong = namedtuple('LatLong', 'lat long')


async def extract_latlongs(network):
    """
    Performs a trace that extracts latitudes and longitudes from
    all equipment in the network.
    `network` The network to trace. Will trace the entire network from the primary sources.
    Returns Tuple of a list of latitudes and a list of longitudes,
    """
    latlongs = []

    async def save_points(equip, stopping):
        for point in equip.position_points:
            latlongs.append(LatLong(lat=point.latitude, long=point.longitude))

    start_item = network.get_primary_sources()[0]
    trace = Traversal(queue_next=queue_next_equipment, start_item=start_item, process_queue=LifoQueue(), step_actions=[save_points])
    await trace.trace()
    return latlongs





#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from typing import Optional

from zepben.evolve.model.cim.iec61970.base.meas.control import Control
from zepben.evolve.model.cim.iec61970.base.scada.remote_point import RemotePoint

__all__ = ["RemoteControl"]


class RemoteControl(RemotePoint):
    """
    Remote controls are outputs that are sent by the remote unit to actuators in the process.
    """

    control: Optional[Control] = None
    """The `zepben.evolve.iec61970.base.meas.control.Control` for the `RemoteControl` point."""



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations
import zepben.evolve.services.common.resolver as resolver

from zepben.protobuf.cim.iec61968.common.Agreement_pb2 import Agreement as PBAgreement
from zepben.protobuf.cim.iec61968.customers.CustomerAgreement_pb2 import CustomerAgreement as PBCustomerAgreement
from zepben.protobuf.cim.iec61968.customers.CustomerKind_pb2 import CustomerKind as PBCustomerKind
from zepben.protobuf.cim.iec61968.customers.Customer_pb2 import Customer as PBCustomer
from zepben.protobuf.cim.iec61968.customers.PricingStructure_pb2 import PricingStructure as PBPricingStructure
from zepben.protobuf.cim.iec61968.customers.Tariff_pb2 import Tariff as PBTariff

from zepben.evolve.model import Customer
from zepben.evolve.services.customer.customers import CustomerService
from zepben.evolve.services.common import *
from zepben.evolve.model import Agreement
from zepben.evolve.model import CustomerAgreement, CustomerKind, PricingStructure, Tariff

__all__ = ["agreement_to_cim", "tariff_to_cim", "customer_to_cim", "customeragreement_to_cim", "pricingstructure_to_cim"]


def agreement_to_cim(pb: PBAgreement, cim: Agreement, service: BaseService):
    document_to_cim(pb.doc, cim, service)


def customer_to_cim(pb: PBCustomer, service: CustomerService):
    cim = Customer(mrid=pb.mrid(), kind=CustomerKind[PBCustomerKind.Name(pb.kind)])
    for mrid in pb.customerAgreementMRIDs:
        service.resolve_or_defer_reference(resolver.agreements(cim), mrid)
    organisationrole_to_cim(getattr(pb, 'or'), cim)
    service.add(cim)


def customeragreement_to_cim(pb: PBCustomerAgreement, service: CustomerService):
    cim = CustomerAgreement(mrid=pb.mrid())
    service.resolve_or_defer_reference(resolver.customer(cim), pb.customerMRID)
    for mrid in pb.pricingStructureMRIDs:
        service.resolve_or_defer_reference(resolver.pricing_structures(cim), mrid)
    agreement_to_cim(pb.agr, cim, service)
    service.add(cim)


def pricingstructure_to_cim(pb: PBPricingStructure, service: CustomerService):
    cim = PricingStructure(mrid=pb.mrid())
    for mrid in pb.tariffMRIDs:
        service.resolve_or_defer_reference(resolver.tariffs(cim), mrid)
    document_to_cim(pb.doc, cim, service)
    service.add(cim)


def tariff_to_cim(self, pb: PBTariff, service: CustomerService):
    cim = Tariff(mrid=pb.mrid())
    document_to_cim(pb.doc, cim, self.service)
    service.add(cim)


PBAgreement.to_cim = agreement_to_cim
PBCustomer.to_cim = customer_to_cim
PBCustomerAgreement.to_cim = customeragreement_to_cim
PBPricingStructure.to_cim = pricingstructure_to_cim
PBTariff.to_cim = PBTariff



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.


from __future__ import annotations
from typing import Optional, List, Generator

__all__ = ["Loop"]

from zepben.evolve.model.cim.iec61970.base.core.identified_object import IdentifiedObject
from zepben.evolve.util import safe_remove, ngen, nlen, get_by_mrid


class Loop(IdentifiedObject):
    """Missing description"""

    loop: Optional[Loop] = None
    _circuits: Optional[List[Circuit]] = None
    _substations: Optional[List[Substation]] = None
    _energizing_substations: Optional[List[Substation]] = None

    def __init__(self, circuits: List[Circuit] = None, substations: List[Substation] = None, energizing_substations: List[Substation] = None):
        if circuits:
            for term in circuits:
                self.add_circuit(term)

        if substations:
            for sub in substations:
                self.add_substation(sub)

        if substations:
            for sub in energizing_substations:
                self.add_energizing_substation(sub)

    @property
    def circuits(self) -> Generator[Circuit, None, None]:
        """
        Sub-transmission `zepben.evolve.cim.infiec61970.base.core.circuit.Circuit`s that form part of this loop.
        """
        return ngen(self._circuits)

    @property
    def substations(self) -> Generator[Substation, None, None]:
        """
        The `zepben.evolve.cim.iec61970.base.core.substation.Substation`s that are powered by this `Loop`.
        """
        return ngen(self._substations)

    @property
    def energizing_substations(self) -> Generator[Substation, None, None]:
        """
        The `zepben.evolve.cim.iec61970.base.core.substation.Substation`s that normally energize this `Loop`.
        """
        return ngen(self._energizing_substations)

    def num_circuits(self):
        """Return the number of end `zepben.evolve.cim.infiec61970.base.core.circuit.Circuit`s associated with this `Loop`"""
        return nlen(self._circuits)

    def get_circuit(self, mrid: str) -> Loop:
        """
        Get the `zepben.evolve.cim.infiec61970.base.core.circuit.Circuit` for this `Loop` identified by `mrid`

        `mrid` the mRID of the required `zepben.evolve.cim.infiec61970.base.core.circuit.Circuit`
        Returns The `zepben.evolve.cim.infiec61970.base.core.circuit.Circuit` with the specified `mrid` if it exists
        Raises `KeyError` if `mrid` wasn't present.
        """
        return get_by_mrid(self._circuits, mrid)

    def add_circuit(self, circuit: Circuit) -> Loop:
        """
        Associate an `zepben.evolve.cim.infiec61970.base.core.circuit.Circuit` with this `Loop`

        `circuit` the `zepben.evolve.cim.infiec61970.base.core.circuit.Circuit` to associate with this `Loop`.
        Returns A reference to this `Loop` to allow fluent use.
        Raises `ValueError` if another `Circuit` with the same `mrid` already exists for this `Loop`.
        """
        if self._validate_reference(circuit, self.get_circuit, "An Circuit"):
            return self
        self._circuits = list() if self._circuits is None else self._circuits
        self._circuits.append(circuit)
        return self

    def remove_circuits(self, circuit: Circuit) -> Loop:
        """
        Disassociate `circuit` from this `Loop`

        `circuit` the `zepben.evolve.cim.infiec61970.base.core.circuit.Circuit` to disassociate from this `Loop`.
        Returns A reference to this `Loop` to allow fluent use.
        Raises `ValueError` if `circuit` was not associated with this `Loop`.
        """
        self._circuits = safe_remove(self._circuits, circuit)
        return self

    def clear_circuits(self) -> Loop:
        """
        Clear all end circuits.
        Returns A reference to this `Loop` to allow fluent use.
        """
        self._circuits = None
        return self

    def num_substations(self):
        """Return the number of end `zepben.evolve.cim.iec61970.base.core.substation.Substation`s associated with this `Loop`"""
        return nlen(self._substations)

    def get_substation(self, mrid: str) -> Loop:
        """
        Get the `zepben.evolve.cim.iec61970.base.core.substation.Substation` for this `Loop` identified by `mrid`

        `mrid` the mRID of the required `zepben.evolve.cim.iec61970.base.core.substation.Substation`
        Returns The `zepben.evolve.cim.iec61970.base.core.substation.Substation` with the specified `mrid` if it exists
        Raises `KeyError` if `mrid` wasn't present.
        """
        return get_by_mrid(self._substations, mrid)

    def add_substation(self, substation: Substation) -> Loop:
        """
        Associate an `zepben.evolve.cim.iec61970.base.core.substation.Substation` with this `Loop`

        `substation` the `zepben.evolve.cim.iec61970.base.core.substation.Substation` to associate with this `Loop`.
        Returns A reference to this `Loop` to allow fluent use.
        Raises `ValueError` if another `Substation` with the same `mrid` already exists for this `Loop`.
        """
        if self._validate_reference(substation, self.get_substation, "An Substation"):
            return self
        self._substations = list() if self._substations is None else self._substations
        self._substations.append(substation)
        return self

    def remove_substations(self, substation: Substation) -> Loop:
        """
        Disassociate `substation` from this `Loop`

        `substation` the `zepben.evolve.cim.iec61970.base.core.substation.Substation` to disassociate from this `Loop`.
        Returns A reference to this `Loop` to allow fluent use.
        Raises `ValueError` if `substation` was not associated with this `Loop`.
        """
        self._substations = safe_remove(self._substations, substation)
        return self

    def clear_substations(self) -> Loop:
        """
        Clear all end substations.
        Returns A reference to this `Loop` to allow fluent use.
        """
        self._substations = None
        return self

    def num_energizing_substations(self):
        """Return the number of end `zepben.evolve.cim.iec61970.base.core.substation.Substation`s associated with this `Loop`"""
        return nlen(self._energizing_substations)

    def get_energizing_substation(self, mrid: str) -> Loop:
        """
        Get the `zepben.evolve.cim.iec61970.base.core.substation.Substation` for this `Loop` identified by `mrid`

        `mrid` the mRID of the required `zepben.evolve.cim.iec61970.base.core.substation.Substation`
        Returns The `zepben.evolve.cim.iec61970.base.core.substation.Substation` with the specified `mrid` if it exists
        Raises `KeyError` if `mrid` wasn't present.
        """
        return get_by_mrid(self._energizing_substations, mrid)

    def add_energizing_substation(self, substation: Substation) -> Loop:
        """
        Associate an `zepben.evolve.cim.iec61970.base.core.substation.Substation` with this `Loop`

        `substation` the `zepben.evolve.cim.iec61970.base.core.substation.Substation` to associate with this `Loop`.
        Returns A reference to this `Loop` to allow fluent use.
        Raises `ValueError` if another `Substation` with the same `mrid` already exists for this `Loop`.
        """
        if self._validate_reference(substation, self.get_energizing_substation, "An Substation"):
            return self
        self._energizing_substations = list() if self._energizing_substations is None else self._energizing_substations
        self._energizing_substations.append(substation)
        return self

    def remove_energizing_substations(self, substation: Substation) -> Loop:
        """
        Disassociate `substation` from this `Loop`

        `substation` the `zepben.evolve.cim.iec61970.base.core.substation.Substation` to disassociate from this `Loop`.
        Returns A reference to this `Loop` to allow fluent use.
        Raises `ValueError` if `substation` was not associated with this `Loop`.
        """
        self._energizing_substations = safe_remove(self._energizing_substations, substation)
        return self

    def clear_energizing_substations(self) -> Loop:
        """
        Clear all end energizing_substations.
        Returns A reference to this `Loop` to allow fluent use.
        """
        self._energizing_substations = None
        return self



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

import logging
from abc import ABCMeta
from dataclassy import dataclass
from typing import Union, Callable, Any
from uuid import UUID


from zepben.evolve.util import require, CopyableUUID

__all__ = ["IdentifiedObject"]

logger = logging.getLogger(__name__)


@dataclass(slots=True)
class IdentifiedObject(object, metaclass=ABCMeta):
    """
    Root class to provide common identification for all classes needing identification and naming attributes.
    Everything should extend this class, however it's not mandated that every subclass must use all the fields
    defined here.

    All names of attributes of classes extending this class *must* directly reflect CIM properties if they have a direct
    relation, however must be in snake case to keep the phases PEP compliant.
    """

    mrid: Union[str, UUID] = CopyableUUID()
    """Master resource identifier issued by a model authority. The mRID is unique within an exchange context. 
    Global uniqueness is easily achieved by using a UUID, as specified in RFC 4122, for the mRID. The use of UUID is strongly recommended."""

    name: str = ""
    """The name is any free human readable and possibly non unique text naming the object."""

    description: str = ""
    """a free human readable text describing or naming the object. It may be non unique and may not correlate to a naming hierarchy."""

    def __str__(self):
        return f"{self.__class__.__name__}{{{'|'.join(a for a in (str(self.mrid), self.name) if a)}}}"

    def _validate_reference(self, other: IdentifiedObject, getter: Callable[[str], IdentifiedObject], type_descr: str) -> bool:
        """
        Validate whether a given reference exists to `other` using the provided getter function.

        `other` The object to look up with the getter using its mRID.
        `getter` A function that takes an mRID and returns an object if it exists, and throws a ``KeyError`` if it couldn't be found.
        `type_descr` The type description to use for the lazily generated error message. Should be of the form "A[n] type(other)"
        Returns True if `other` was retrieved with `getter` and was equivalent, False otherwise.
        Raises `ValueError` if the object retrieved from `getter` is not `other`.
        """
        try:
            get_result = getter(other.mrid)
            require(get_result is other, lambda: f"{type_descr} with mRID {other.mrid} already exists in {str(self)}")
            return True
        except (KeyError, AttributeError):
            return False

    def _validate_reference_by_sn(self, field: Any, other: IdentifiedObject, getter: Callable[[Any], IdentifiedObject], type_descr: str, field_name: str = "sequence_number") -> bool:
        """
        Validate whether a given reference exists to `other` using the provided getter function called with `field`.

        `other` The object to compare against.
        `getter` A function that takes `field` and returns an `IdentifiedObject` if it exists, and throws an `IndexError` if it couldn't be found.
        `type_descr` The type description to use for the lazily generated error message. Should be of the form "A[n] type(other)"
        Returns True if `other` was retrieved with a call to `getter(field)` and was equivalent, False otherwise.
        Raises `ValueError` if an object is retrieved from `getter` and it is not `other`.
        """
        try:
            get_result = getter(field)
            require(get_result is other, lambda: f"{type_descr} with {field_name} {field} already exists in {str(self)}")
            return True
        except IndexError:
            return False






#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

from typing import Optional

from zepben.evolve.model.cim.iec61970.base.meas.iopoint import IoPoint

__all__ = ["Control"]


class Control(IoPoint):
    """
    Control is used for supervisory/device control. It represents control outputs that are used to change the state in a
    process, e.g. close or open breaker, a set point value or a raise lower command.
    """

    power_system_resource_mrid: Optional[str] = None
    """AnalogValue represents an analog MeasurementValue."""
    
    remote_control: Optional[RemoteControl] = None
    """AnalogValue represents an analog MeasurementValue."""



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

from abc import abstractmethod
from typing import Iterable, Dict, Optional, Set, Tuple

from dataclassy import dataclass


from zepben.evolve.streaming.exceptions import UnsupportedOperationException
from zepben.protobuf.nc.nc_data_pb2 import NetworkIdentifiedObject

from zepben.evolve.streaming.grpc import GrpcClient, GrpcResult

__all__ = ["CimConsumerClient", "MultiObjectResult", "extract_identified_object"]


@dataclass()
class MultiObjectResult(object):
    value: Dict[str, IdentifiedObject]
    failed: Set[str]


class CimConsumerClient(GrpcClient):

    @abstractmethod
    async def get_identified_object(self, service: BaseService, mrid: str) -> GrpcResult:
        """
        Retrieve the object with the given `mrid` and store the result in the `service`.
                                                                                                                 
        Exceptions that occur during sending will be caught and passed to all error handlers that have been registered against this client.

        Returns a `GrpcResult` with a result of one of the following:
             - The object if found
             - null if an object could not be found or it was found but not added to `service` (see `zepben.evolve.common.base_service.BaseService.add`).
             - An `Exception` if an error occurred while retrieving or processing the object, in which case, `GrpcResult.was_successful` will return false.
        """
        raise NotImplementedError()

    @abstractmethod
    async def get_identified_objects(self, service: BaseService, mrids: Iterable[str]) -> GrpcResult:
        """
        Retrieve the objects with the given `mrids` and store the results in the `service`.

        Exceptions that occur during sending will be caught and passed to all error handlers that have been registered against this client.

        WARNING: This operation is not atomic upon `service`, and thus if processing fails partway through `mrids`, any previously successful mRID will have been
        added to the service, and thus you may have an incomplete `BaseService`. Also note that adding to the `service` may not occur for an object if another
        object with the same mRID is already present in `service`. `MultiObjectResult.failed` can be used to check for mRIDs that were retrieved but not
        added to `service`.
        
        Returns a `GrpcResult` with a result of one of the following:
        - A `MultiObjectResult` containing a map of the retrieved objects keyed by mRID. If an item is not found it will be excluded from the map.
          If an item couldn't be added to `service` its mRID will be present in `MultiObjectResult.failed` (see `zepben.evolve.common.base_service.BaseService.add`).
        - An `Exception` if an error occurred while retrieving or processing the objects, in which case, `GrpcResult.was_successful` will return false.
          Note the warning above in this case.
        """
        raise NotImplementedError()


def extract_identified_object(service: NetworkService, nio: NetworkIdentifiedObject) -> Tuple[Optional[IdentifiedObject], str]:
    """
    Add an equipment to the network.
    `stub` A network consumer stub.
    `network` The network to add the equipment to.
    `equipment_io` The equipment identified object returned by the server.
    Raises `UnsupportedOperationException` if `nio` was invalid/unset.
    """
    io_type = nio.WhichOneof("identifiedObject")
    if io_type:
        pbio = getattr(nio, io_type)
        return service.add_from_pb(pbio), pbio.mrid()
    else:
        raise UnsupportedOperationException(f"Received a NetworkIdentifiedObject where no field was set")





#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from zepben.protobuf.cim.iec61970.base.meas.MeasurementValue_pb2 import MeasurementValue as PBMeasurementValue
from zepben.protobuf.cim.iec61970.base.meas.AccumulatorValue_pb2 import AccumulatorValue as PBAccumulatorValue
from zepben.protobuf.cim.iec61970.base.meas.AnalogValue_pb2 import AnalogValue as PBAnalogValue
from zepben.protobuf.cim.iec61970.base.meas.DiscreteValue_pb2 import DiscreteValue as PBDiscreteValue
from zepben.evolve.model import AnalogValue, AccumulatorValue, DiscreteValue, MeasurementValue

from zepben.evolve.services.measurement import MeasurementService

__all__ = ["measurementvalue_to_cim", "analogvalue_to_cim", "accumulatorvalue_to_cim", "discretevalue_to_cim"]


# IEC61970 MEAS
def measurementvalue_to_cim(pb: PBMeasurementValue, cim: MeasurementValue):
    cim.time_stamp = pb.timeStamp.ToDatetime()


def analogvalue_to_cim(pb: PBAnalogValue, service: MeasurementService):
    cim = AnalogValue(analog_mrid=pb.analogMRID, value=pb.value)
    measurementvalue_to_cim(pb.mv, cim)
    service.add(cim)


def accumulatorvalue_to_cim(pb: PBAccumulatorValue, service: MeasurementService):
    cim = AccumulatorValue(accumulator_mrid=pb.accumulatorMRID, value=pb.value)
    measurementvalue_to_cim(pb.mv, cim)
    service.add(cim)


def discretevalue_to_cim(pb: PBDiscreteValue, service: MeasurementService):
    cim = DiscreteValue(discrete_mrid=pb.discreteMRID, value=pb.value)
    measurementvalue_to_cim(pb.mv, cim)
    service.add(cim)


PBAccumulatorValue.to_cim = accumulatorvalue_to_cim
PBAnalogValue.to_cim = analogvalue_to_cim
PBDiscreteValue.to_cim = discretevalue_to_cim
PBMeasurementValue.to_cim = measurementvalue_to_cim


#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
import pytest
from hypothesis import given, settings, Phase
from zepben.protobuf.nc.nc_data_pb2 import NetworkIdentifiedObject, IdentifiedObjectGroup
from zepben.protobuf.nc.nc_responses_pb2 import GetIdentifiedObjectsResponse
from unittest.mock import MagicMock

from test.pb_creators import networkidentifiedobjects, aclinesegment
from zepben.evolve import NetworkConsumerClient, NetworkService


class TestConsumer(object):

    @pytest.mark.asyncio
    @given(networkidentifiedobjects())
    @settings(max_examples=1, phases=(Phase.explicit, Phase.reuse, Phase.generate))
    async def test_retrieve_supported_types(self, networkidentifiedobjects):
        network_service = NetworkService()
        for nio in networkidentifiedobjects:
            response = GetIdentifiedObjectsResponse(objectGroup=IdentifiedObjectGroup(identifiedObject=nio))
            stub = MagicMock(**{"getIdentifiedObjects.return_value": [response]})
            client = NetworkConsumerClient(stub=stub)
            pbio = getattr(nio, nio.WhichOneof("identifiedObject"), None)
            result = await client.get_identified_objects(network_service, [pbio.mrid()])
            assert result.was_successful
            if pbio.mrid():
                assert result.result.value[pbio.mrid()] is not None, f"type: {nio.WhichOneof('identifiedObject')} mrid: {pbio.mrid()}"
                assert network_service.get(pbio.mrid()) is result.result.value[pbio.mrid()]
            else:
                assert pbio.mrid() in result.result.failed

    @pytest.mark.asyncio
    @given(aclinesegment())
    async def test_get_identifiedobject(self, aclinesegment):
        network_service = NetworkService()
        nio = NetworkIdentifiedObject(acLineSegment=aclinesegment)
        response = GetIdentifiedObjectsResponse(objectGroup=IdentifiedObjectGroup(identifiedObject=nio))
        stub = MagicMock(**{"getIdentifiedObjects.return_value": [response]})
        client = NetworkConsumerClient(stub=stub)
        pbio = getattr(nio, nio.WhichOneof("identifiedObject"), None)
        result = await client.get_identified_object(network_service, pbio.mrid())
        assert result.was_successful
        assert result.result is not None
        assert result.result.mrid == pbio.mrid()

        stub = MagicMock(**{"getIdentifiedObjects.return_value": []})
        client = NetworkConsumerClient(stub=stub)
        pbio = getattr(nio, nio.WhichOneof("identifiedObject"), None)
        result = await client.get_identified_object(network_service, "fakemrid")
        assert result.was_successful
        assert result.result is None






#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

from typing import Optional, Generator, List

from zepben.evolve.model.cim.iec61970.base.core.equipment_container import EquipmentContainer
from zepben.evolve.model.cim.iec61970.base.core.regions import SubGeographicalRegion
from zepben.evolve.util import nlen, get_by_mrid, ngen, safe_remove

__all__ = ["Substation"]


class Substation(EquipmentContainer):
    """
    A collection of equipment for purposes other than generation or utilization, through which electric energy in bulk
    is passed for the purposes of switching or modifying its characteristics.
    """

    sub_geographical_region: Optional[SubGeographicalRegion] = None
    """The SubGeographicalRegion containing the substation."""

    _normal_energized_feeders: Optional[List[Feeder]] = None

    _loops: Optional[List[Loop]] = None

    _energized_loops: Optional[List[Loop]] = None

    _circuits: Optional[List[Circuit]] = None

    def __init__(self, equipment: List[Equipment] = None, normal_energized_feeders: List[Feeder] = None, loops: List[Loop] = None,
                 energized_loops: List[Loop] = None, circuits: List[Circuit] = None):
        super(Substation, self).__init__(equipment)
        if normal_energized_feeders:
            for feeder in normal_energized_feeders:
                self.add_feeder(feeder)
        if loops:
            for loop in loops:
                self.add_loop(loop)
        if energized_loops:
            for loop in energized_loops:
                self.add_energized_loop(loop)
        if circuits:
            for circuit in circuits:
                self.add_circuit(circuit)

    @property
    def circuits(self) -> Generator[Circuit, None, None]:
        """
        The `zepben.evolve.cim.infiec61970.feeder.circuit.Circuit`s originating from this substation.
        """
        return ngen(self._circuits)

    @property
    def loops(self) -> Generator[Loop, None, None]:
        """
        The `zepben.evolve.cim.infiec61970.feeder.loop.Loop` originating from this substation.
        """
        return ngen(self._loops)

    @property
    def energized_loops(self) -> Generator[Loop, None, None]:
        """
        The `zepben.evolve.cim.infiec61970.feeder.loop.Loop`s originating from this substation that are energised.
        """
        return ngen(self._energized_loops)

    @property
    def feeders(self) -> Generator[Feeder, None, None]:
        """
        The normal energized feeders of the substation. Also used for naming purposes.
        """
        return ngen(self._normal_energized_feeders)

    def num_feeders(self):
        """
        Returns The number of `zepben.evolve.cim.iec61970.base.core.equipment_container.Feeder`s associated with this `Substation`
        """
        return nlen(self._normal_energized_feeders)

    def get_feeder(self, mrid: str) -> Substation:
        """
        Get the `zepben.evolve.cim.iec61970.base.core.equipment_container.Feeder` for this `Substation` identified by `mrid`

        `mrid` The mRID of the required `zepben.evolve.cim.iec61970.base.core.equipment_container.Feeder`
        Returns The `zepben.evolve.cim.iec61970.base.core.equipment_container.Feeder` with the specified `mrid` if it exists
        Raises `KeyError` if `mrid` wasn't present.
        """
        return get_by_mrid(self._normal_energized_feeders, mrid)

    def add_feeder(self, feeder: Feeder) -> Substation:
        """
        Associate a `zepben.evolve.cim.iec61970.base.core.equipment_container.Feeder` with this `Substation`

        `feeder` The `zepben.evolve.cim.iec61970.base.core.equipment_container.Feeder` to associate with this `Substation`.
        Returns A reference to this `Substation` to allow fluent use.
        Raises `ValueError` if another `Feeder` with the same `mrid` already exists for this `Substation`.
        """
        if self._validate_reference(feeder, self.get_feeder, "A Feeder"):
            return self
        self._normal_energized_feeders = list() if self._normal_energized_feeders is None else self._normal_energized_feeders
        self._normal_energized_feeders.append(feeder)
        return self

    def remove_feeder(self, feeder: Feeder) -> Substation:
        """
        Disassociate `feeder` from this `Substation`

        `feeder` The `zepben.evolve.cim.iec61970.base.core.equipment_container.Feeder` to disassociate from this `Substation`.
        Returns A reference to this `Substation` to allow fluent use.
        Raises `ValueError` if `feeder` was not associated with this `Substation`.
        """
        self._normal_energized_feeders = safe_remove(self._normal_energized_feeders, feeder)
        return self

    def clear_feeders(self) -> Substation:
        """
        Clear all current `Feeder`s.
        Returns A reference to this `Substation` to allow fluent use.
        """
        self._normal_energized_feeders = None
        return self

    def num_loops(self):
        """
        Returns The number of `zepben.evolve.cim.infiec61970.feeder.loop.Loop`s associated with this `Substation`
        """
        return nlen(self._loops)

    def get_loop(self, mrid: str) -> Substation:
        """
        Get the `zepben.evolve.cim.infiec61970.feeder.loop.Loop` for this `Substation` identified by `mrid`

        `mrid` The mRID of the required `zepben.evolve.cim.infiec61970.feeder.loop.Loop`
        Returns The `zepben.evolve.cim.infiec61970.feeder.loop.Loop` with the specified `mrid` if it exists
        Raises `KeyError` if `mrid` wasn't present.
        """
        return get_by_mrid(self._loops, mrid)

    def add_loop(self, loop: Loop) -> Substation:
        """
        Associate a `zepben.evolve.cim.infiec61970.feeder.loop.Loop` with this `Substation`

        `loop` The `zepben.evolve.cim.infiec61970.feeder.loop.Loop` to associate with this `Substation`.
        Returns A reference to this `Substation` to allow fluent use.
        Raises `ValueError` if another `Loop` with the same `mrid` already exists for this `Substation`.
        """
        if self._validate_reference(loop, self.get_loop, "A Loop"):
            return self
        self._loops = list() if self._loops is None else self._loops
        self._loops.append(loop)
        return self

    def remove_loop(self, loop: Loop) -> Substation:
        """
        Disassociate `loop` from this `Substation`

        `loop` The `zepben.evolve.cim.infiec61970.feeder.loop.Loop` to disassociate from this `Substation`.
        Returns A reference to this `Substation` to allow fluent use.
        Raises `ValueError` if `loop` was not associated with this `Substation`.
        """
        self._loops = safe_remove(self._loops, loop)
        return self

    def clear_loops(self) -> Substation:
        """
        Clear all current `Loop`s.
        Returns A reference to this `Substation` to allow fluent use.
        """
        self._loops = None
        return self

    def num_energized_loops(self):
        """
        Returns The number of `zepben.evolve.cim.infiec61970.feeder.loop.Loop`s associated with this `Substation`
        """
        return nlen(self._energized_loops)

    def get_energized_loop(self, mrid: str) -> Substation:
        """
        Get the `zepben.evolve.cim.infiec61970.feeder.loop.Loop` for this `Substation` identified by `mrid`

        `mrid` The mRID of the required `zepben.evolve.cim.infiec61970.feeder.loop.Loop`
        Returns The `zepben.evolve.cim.infiec61970.feeder.loop.Loop` with the specified `mrid` if it exists
        Raises `KeyError` if `mrid` wasn't present.
        """
        return get_by_mrid(self._energized_loops, mrid)

    def add_energized_loop(self, loop: Loop) -> Substation:
        """
        Associate a `zepben.evolve.cim.infiec61970.feeder.loop.Loop` with this `Substation`

        `loop` The `zepben.evolve.cim.infiec61970.feeder.loop.Loop` to associate with this `Substation`.
        Returns A reference to this `Substation` to allow fluent use.
        Raises `ValueError` if another `Loop` with the same `mrid` already exists for this `Substation`.
        """
        if self._validate_reference(loop, self.get_energized_loop, "A Loop"):
            return self
        self._energized_loops = list() if self._energized_loops is None else self._energized_loops
        self._energized_loops.append(loop)
        return self

    def remove_energized_loop(self, loop: Loop) -> Substation:
        """
        Disassociate `loop` from this `Substation`

        `loop` The `zepben.evolve.cim.infiec61970.feeder.loop.Loop` to disassociate from this `Substation`.
        Returns A reference to this `Substation` to allow fluent use.
        Raises `ValueError` if `loop` was not associated with this `Substation`.
        """
        self._energized_loops = safe_remove(self._energized_loops, loop)
        return self

    def clear_energized_loops(self) -> Substation:
        """
        Clear all current `Loop`s.
        Returns A reference to this `Substation` to allow fluent use.
        """
        self._energized_loops = None
        return self

    def num_circuits(self):
        """
        Returns The number of `zepben.evolve.cim.infiec61970.feeder.circuit.Circuit`s associated with this `Substation`
        """
        return nlen(self._circuits)

    def get_circuit(self, mrid: str) -> Substation:
        """
        Get the `zepben.evolve.cim.infiec61970.feeder.circuit.Circuit` for this `Substation` identified by `mrid`

        `mrid` The mRID of the required `zepben.evolve.cim.infiec61970.feeder.circuit.Circuit`
        Returns The `zepben.evolve.cim.infiec61970.feeder.circuit.Circuit` with the specified `mrid` if it exists
        Raises `KeyError` if `mrid` wasn't present.
        """
        return get_by_mrid(self._circuits, mrid)

    def add_circuit(self, circuit: Circuit) -> Substation:
        """
        Associate a `zepben.evolve.cim.infiec61970.feeder.circuit.Circuit` with this `Substation`

        `circuit` The `zepben.evolve.cim.infiec61970.feeder.circuit.Circuit` to associate with this `Substation`.
        Returns A reference to this `Substation` to allow fluent use.
        Raises `ValueError` if another `Circuit` with the same `mrid` already exists for this `Substation`.
        """
        if self._validate_reference(circuit, self.get_circuit, "A Circuit"):
            return self
        self._circuits = list() if self._circuits is None else self._circuits
        self._circuits.append(circuit)
        return self

    def remove_circuit(self, circuit: Circuit) -> Substation:
        """
        Disassociate `circuit` from this `Substation`

        `circuit` The `zepben.evolve.cim.infiec61970.feeder.circuit.Circuit` to disassociate from this `Substation`.
        Returns A reference to this `Substation` to allow fluent use.
        Raises `ValueError` if `circuit` was not associated with this `Substation`.
        """
        self._circuits = safe_remove(self._circuits, circuit)
        return self

    def clear_circuits(self) -> Substation:
        """
        Clear all current `Circuit`s.
        Returns A reference to this `Substation` to allow fluent use.
        """
        self._circuits = None
        return self



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
from abc import ABCMeta
from typing import Optional

from dataclassy import dataclass


from zepben.evolve.services.common.base_service import BaseService
from zepben.protobuf.cim.iec61970.base.core.IdentifiedObject_pb2 import IdentifiedObject as PBIdentifiedObject
from zepben.protobuf.cim.iec61968.common.Document_pb2 import Document as PBDocument

from zepben.evolve.model.cim.iec61968.common.document import Document
from zepben.evolve.model import Organisation
from zepben.evolve.model import OrganisationRole
from zepben.protobuf.cim.iec61968.common.Organisation_pb2 import Organisation as PBOrganisation
from zepben.protobuf.cim.iec61968.common.OrganisationRole_pb2 import OrganisationRole as PBOrganisationRole
from zepben.evolve.model.cim.iec61970.base.core.identified_object import IdentifiedObject
from zepben.evolve.services.common import resolver

__all__ = ["identifiedobject_to_cim", "document_to_cim", "organisation_to_cim", "organisationrole_to_cim", "BaseProtoToCim"]


# IEC61970 CORE #
def identifiedobject_to_cim(pb: PBIdentifiedObject, cim: IdentifiedObject, service: BaseService):
    cim.mrid = pb.mRID
    cim.name = pb.name
    cim.description = pb.description


# IEC61968 COMMON #
def document_to_cim(pb: PBDocument, cim: Document, service: BaseService):
    cim.title = pb.title
    cim.created_date_time = pb.createdDateTime.ToDatetime()
    cim.author_name = pb.authorName
    cim.type = pb.type
    cim.status = pb.status
    cim.comment = pb.comment
    identifiedobject_to_cim(pb.io, cim, service)


def organisation_to_cim(pb: PBOrganisation, service: BaseService) -> Optional[Organisation]:
    cim = Organisation()
    identifiedobject_to_cim(pb.io, cim, service)
    return cim if service.add(cim) else None


def organisationrole_to_cim(pb: PBOrganisationRole, cim: OrganisationRole, service: BaseService):
    cim.organisation = service.resolve_or_defer_reference(resolver.organisation(cim), pb.organisationMRID)
    identifiedobject_to_cim(pb.io, cim, service)


PBDocument.to_cim = document_to_cim
PBOrganisation.to_cim = organisation_to_cim
PBOrganisationRole.to_cim = organisationrole_to_cim
PBIdentifiedObject.to_cim = identifiedobject_to_cim


@dataclass(slots=True)
class BaseProtoToCim(object, metaclass=ABCMeta):
    service: BaseService




#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

from typing import Optional

from zepben.evolve.model.cim.iec61970.base.core.identified_object import IdentifiedObject

__all__ = ['PowerSystemResource']


class PowerSystemResource(IdentifiedObject):
    """
    Abstract class, should only be used through subclasses.
    A power system resource can be an item of equipment such as a switch, an equipment container containing many individual
    items of equipment such as a substation, or an organisational entity such as sub-control area. Power system resources
    can have measurements associated.
    """

    location: Optional[Location] = None
    """A `zepben.evolve.iec61968.common.location.Location` for this resource."""

    asset_info: Optional[AssetInfo] = None
    """A subclass of `zepben.evolve.iec61968.assets.asset_info.AssetInfo` providing information about the asset associated with this PowerSystemResource."""






#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

from typing import Optional, Generator, List

from zepben.evolve.model.cim.iec61968.common.document import Document
from zepben.evolve.util import get_by_mrid, nlen, ngen, safe_remove

__all__ = ["PricingStructure"]


class PricingStructure(Document):
    """
    Grouping of pricing components and prices used in the creation of customer charges and the eligibility
    criteria under which these terms may be offered to a customer. The reasons for grouping include state,
    customer classification, site characteristics, classification (i.e. fee price structure, deposit price
    structure, electric service price structure, etc.) and accounting requirements.
    """
    _tariffs: Optional[List[Tariff]] = None

    def __init__(self, tariffs: List[Tariff] = None):
        if tariffs:
            for tariff in tariffs:
                self.add_tariff(tariff)

    def num_tariffs(self):
        """
        Returns The number of `zepben.evolve.cim.iec61968.customers.tariff.Tariff`s associated with this `PricingStructure`
        """
        return nlen(self._tariffs)

    @property
    def tariffs(self) -> Generator[Tariff, None, None]:
        """
        The `zepben.evolve.cim.iec61968.customers.tariff.Tariff`s of this `PricingStructure`.
        """
        return ngen(self._tariffs)

    def get_tariff(self, mrid: str) -> Tariff:
        """
        Get the `zepben.evolve.cim.iec61968.customers.tariff.Tariff` for this `PricingStructure` identified by `mrid`

        `mrid` the mRID of the required `zepben.evolve.cim.iec61968.customers.tariff.Tariff`
        Returns The `zepben.evolve.cim.iec61968.customers.tariff.Tariff` with the specified `mrid` if it exists
        Raises `KeyError` if `mrid` wasn't present.
        """
        return get_by_mrid(self._tariffs, mrid)

    def add_tariff(self, tariff: Tariff) -> PricingStructure:
        """
        Associate a `zepben.evolve.cim.iec61968.customers.tariff.Tariff` with this `PricingStructure`.

        `tariff` the `zepben.evolve.cim.iec61968.customers.tariff.Tariff` to associate with this `PricingStructure`.
        Returns A reference to this `PricingStructure` to allow fluent use.
        Raises `ValueError` if another `Tariff` with the same `mrid` already exists for this `PricingStructure`.
        """
        if self._validate_reference(tariff, self.get_tariff, "A Tariff"):
            return self
        self._tariffs = list() if self._tariffs is None else self._tariffs
        self._tariffs.append(tariff)
        return self

    def remove_tariff(self, tariff: Tariff) -> PricingStructure:
        """
        Disassociate `tariff` from this `PricingStructure`.

        `tariff` the `zepben.evolve.cim.iec61968.customers.tariff.Tariff` to disassociate from this `PricingStructure`.
        Returns A reference to this `PricingStructure` to allow fluent use.
        Raises `ValueError` if `tariff` was not associated with this `PricingStructure`.
        """
        self._tariffs = safe_remove(self._tariffs, tariff)
        return self

    def clear_tariffs(self) -> PricingStructure:
        """
        Clear all tariffs.
        Returns A reference to this `PricingStructure` to allow fluent use.
        """
        self._tariffs = None
        return self
